# -*- coding: utf-8 -*-
"""Top-level package for pytorch-fast-elmo."""

__author__ = """Hunt Zhan"""
__email__ = 'huntzhan.dev@gmail.com'
__version__ = '0.6.12'

# To avoid `undefined symbol` error.
import torch

# pylint: disable=no-name-in-module
from pytorch_stateful_lstm import StatefulUnidirectionalLstm
# from _pytorch_fast_elmo import ElmoCharacterEncoder, ScalarMix
from _pytorch_fast_elmo import ElmoCharacterEncoder

from pytorch_fast_elmo.utils import (
        batch_to_char_ids,
        load_and_build_vocab2id,
        batch_to_word_ids,
)

from pytorch_fast_elmo.factory import (
        ElmoCharacterEncoderFactory,
        ElmoWordEmbeddingFactory,
        ElmoLstmFactory,
        ElmoVocabProjectionFactory,
)

from pytorch_fast_elmo.model import (
        ScalarMix,
        FastElmoBase,
        FastElmo,
        FastElmoWordEmbedding,
        FastElmoPlainEncoder,
        FastElmoWordEmbeddingPlainEncoder,
        FastElmoForwardVocabDistrib,
        FastElmoBackwardVocabDistrib,
        FastElmoWordEmbeddingForwardVocabDistrib,
        FastElmoWordEmbeddingBackwardVocabDistrib,
)



from os.path import dirname, join

import torch
from torch.nn.utils.rnn import pack_padded_sequence, pad_packed_sequence, PackedSequence
import numpy as np
import random

from allennlp.data import Token, Vocabulary, Instance
from allennlp.data.fields import TextField
from allennlp.data.dataset import Batch
from allennlp.nn.util import add_sentence_boundary_token_ids, remove_sentence_boundaries
from allennlp.data.token_indexers.elmo_indexer import (
        ELMoCharacterMapper,
        ELMoTokenCharactersIndexer,
)
from allennlp.modules.elmo import _ElmoCharacterEncoder, _ElmoBiLm, Elmo

from pytorch_fast_elmo import (
        ElmoCharacterEncoder,
        ElmoCharacterEncoderFactory,
        ElmoLstmFactory,
        FastElmo,
        FastElmoWordEmbedding,
        utils,
)

FIXTURES_FODLER = join(dirname(__file__), 'fixtures')
ELMO_OPTIONS_FILE = join(FIXTURES_FODLER, 'options.json')
ELMO_WEIGHT_FILE = join(FIXTURES_FODLER, 'lm_weights.hdf5')

CACHE_VOCAB_FILE = join(FIXTURES_FODLER, 'vocab.txt')
CACHE_EMBD_FILE = join(FIXTURES_FODLER, 'lm_embd.txt')


def test_elmo_character_encoder_simple():
    embedder = ElmoCharacterEncoder(
            char_embedding_cnt=10,
            char_embedding_dim=3,
            filters=[
                    [1, 5],
                    [2, 4],
            ],
            activation='relu',
            num_highway_layers=2,
            output_dim=3,
    )
    inputs = torch.randint(0, 10, (20, 10))
    output = embedder(inputs)
    assert list(output.size()) == [20, 3]


def _sentences_to_ids(sentences):
    indexer = ELMoTokenCharactersIndexer()

    # For each sentence, first create a TextField, then create an instance
    instances = []
    for sentence in sentences:
        tokens = [Token(token) for token in sentence]
        field = TextField(tokens, {'character_ids': indexer})
        instance = Instance({'elmo': field})
        instances.append(instance)

    dataset = Batch(instances)
    vocab = Vocabulary()
    dataset.index_instances(vocab)
    return dataset.as_tensor_dict()['elmo']['character_ids']


def _unpack(tensor, batch_sizes):
    tensor, _ = pad_packed_sequence(
            PackedSequence(tensor, batch_sizes),
            batch_first=True,
    )
    return tensor


def test_elmo_character_encoder_with_allennlp():
    allennlp_embedder = _ElmoCharacterEncoder(
            ELMO_OPTIONS_FILE,
            ELMO_WEIGHT_FILE,
    )
    embedder = ElmoCharacterEncoderFactory(
            ELMO_OPTIONS_FILE,
            ELMO_WEIGHT_FILE,
    ).create()

    allennlp_parameters = [
            '_char_embedding_weights',
            'char_conv_0.bias',
            'char_conv_0.weight',
            'char_conv_1.bias',
            'char_conv_1.weight',
            'char_conv_2.bias',
            'char_conv_2.weight',
            'char_conv_3.bias',
            'char_conv_3.weight',
            'char_conv_4.bias',
            'char_conv_4.weight',
            '_projection.bias',
            '_projection.weight',
    ]
    embedder_parameters = [
            'char_embedding.weight',
            'char_conv_0.bias',
            'char_conv_0.weight',
            'char_conv_1.bias',
            'char_conv_1.weight',
            'char_conv_2.bias',
            'char_conv_2.weight',
            'char_conv_3.bias',
            'char_conv_3.weight',
            'char_conv_4.bias',
            'char_conv_4.weight',
            'output_proj.bias',
            'output_proj.weight',
    ]
    allennlp_parameters_diff = [
            '_highways._layers.0.bias',
            '_highways._layers.0.weight',
            '_highways._layers.1.bias',
            '_highways._layers.1.weight',
    ]
    embedder_parameters_diff = [
            'highway.layers_0.bias',
            'highway.layers_0.weight',
            'highway.layers_1.bias',
            'highway.layers_1.weight',
    ]
    assert len(allennlp_parameters) == len(embedder_parameters)
    assert len(allennlp_parameters_diff) == len(embedder_parameters_diff)

    allennlp_embedder_named_parameters = dict(allennlp_embedder.named_parameters())
    # Same.
    for allennlp_param, embedder_param in zip(allennlp_parameters, embedder_parameters):
        allennlp_w = allennlp_embedder_named_parameters[allennlp_param].data
        embedder_w = embedder.named_parameters()[embedder_param].data

        np.testing.assert_array_equal(embedder_w.numpy(), allennlp_w.numpy())
        assert embedder_w.dtype == allennlp_w.dtype
    # Diff on highway.
    for allennlp_param, embedder_param in zip(allennlp_parameters_diff, embedder_parameters_diff):
        allennlp_w = allennlp_embedder_named_parameters[allennlp_param].data
        embedder_w = embedder.named_parameters()[embedder_param].data

        assert embedder_w.dtype == allennlp_w.dtype
        np.testing.assert_raises(
                AssertionError,
                np.testing.assert_array_equal,
                embedder_w.numpy(),
                allennlp_w.numpy(),
        )

    sentences = [
            ['ELMo', 'helps', 'disambiguate', 'ELMo', 'from', 'Elmo', '.'],
            ['The', 'sentence', '.'],
    ]
    # `(2, 7, 50)`
    character_ids = _sentences_to_ids(sentences)

    # AllenNLP.
    out = allennlp_embedder(character_ids)
    allennlp_token_embedding, _ = remove_sentence_boundaries(out['token_embedding'], out['mask'])
    assert list(allennlp_token_embedding.shape) == [2, 7, 16]

    # Ours.
    inputs = pack_padded_sequence(character_ids, [7, 3], batch_first=True)
    out = embedder(inputs.data)
    ours_token_embedding = _unpack(out, inputs.batch_sizes)
    assert list(ours_token_embedding.shape) == [2, 7, 16]

    np.testing.assert_array_almost_equal(
            ours_token_embedding.data.numpy(),
            allennlp_token_embedding.data.numpy(),
    )


def test_elmo_lstm_factory_simple():
    allennlp_elmo_bilm = _ElmoBiLm(
            ELMO_OPTIONS_FILE,
            ELMO_WEIGHT_FILE,
    )

    embedder = ElmoCharacterEncoderFactory(
            ELMO_OPTIONS_FILE,
            ELMO_WEIGHT_FILE,
    ).create()
    fwd_lstm, bwd_lstm = ElmoLstmFactory(
            ELMO_OPTIONS_FILE,
            ELMO_WEIGHT_FILE,
    ).create(
            enable_forward=True, enable_backward=True)

    sentences_1 = [
            ['ELMo', 'helps', 'disambiguate', 'ELMo', 'from', 'Elmo', '.'],
            ['The', 'sentence', '.'],
    ]
    sentences_2 = [
            ["This", "is", "a", "sentence"],
            ["Here", "'s", "one"],
            ["Another", "one"],
    ]

    # Internal states should be updated.
    for sentences in [sentences_1, sentences_2] * 10:
        # `(2, 7, 50)`
        character_ids = _sentences_to_ids(sentences)

        # AllenNLP.
        allennlp_out = allennlp_elmo_bilm(character_ids)

        # Ours.
        inputs = character_ids
        _beginning_of_sentence_characters = torch.from_numpy(
                np.array(ELMoCharacterMapper.beginning_of_sentence_characters) + 1)
        _end_of_sentence_characters = torch.from_numpy(
                np.array(ELMoCharacterMapper.end_of_sentence_characters) + 1)
        # Add BOS/EOS
        mask = ((inputs > 0).long().sum(dim=-1) > 0).long()
        character_ids_with_bos_eos, mask_with_bos_eos = add_sentence_boundary_token_ids(
                inputs,
                mask,
                _beginning_of_sentence_characters,
                _end_of_sentence_characters,
        )
        # Pack input.
        lengths = mask_with_bos_eos.sum(dim=-1)
        inputs = pack_padded_sequence(character_ids_with_bos_eos, lengths, batch_first=True)
        char_repr = embedder(inputs.data)
        fwd_lstm_hiddens, _ = fwd_lstm(char_repr, inputs.batch_sizes)
        bwd_lstm_hiddens, _ = bwd_lstm(char_repr, inputs.batch_sizes)
        lstm_hiddens = [
                torch.cat([fwd, bwd], dim=-1)
                for fwd, bwd in zip(fwd_lstm_hiddens, bwd_lstm_hiddens)
        ]
        # Unpack output.
        char_repr = _unpack(char_repr, inputs.batch_sizes)
        duplicated_char_repr = torch.cat(
                [char_repr, char_repr],
                dim=-1,
        ) * mask_with_bos_eos.float().unsqueeze(-1)
        lstm_hiddens = [_unpack(hx, inputs.batch_sizes) for hx in lstm_hiddens]

        # TODO: Investigate the numerical stability issue.
        # np.testing.assert_array_almost_equal(
        #         duplicated_char_repr.data.numpy(),
        #         allennlp_out['activations'][0].data.numpy(),
        # )
        # np.testing.assert_array_almost_equal(
        #         lstm_hiddens[0].data.numpy(),
        #         allennlp_out['activations'][1].data.numpy(),
        # )
        np.testing.assert_array_almost_equal(
                lstm_hiddens[1].data.numpy(),
                allennlp_out['activations'][2].data.numpy(),
        )


def test_fast_elmo_with_allennlp():
    fast = FastElmo(
            ELMO_OPTIONS_FILE,
            ELMO_WEIGHT_FILE,
            num_output_representations=2,
            scalar_mix_parameters=[1.0, 1.0, 1.0],
    )

    allennlp = Elmo(
            ELMO_OPTIONS_FILE,
            ELMO_WEIGHT_FILE,
            num_output_representations=2,
            dropout=0.0,
            scalar_mix_parameters=[1.0, 1.0, 1.0],
    )

    sentences_1 = [
            ['ELMo', 'helps', 'disambiguate', 'ELMo', 'from', 'Elmo', '.'],
            ['The', 'sentence', '.'],
    ]
    sentences_2 = [
            ["This", "is", "a", "sentence"],
            ["Here", "'s", "one"],
            ["Another", "one"],
    ]

    for sentences in [sentences_1, sentences_2] * 10:
        random.shuffle(sentences)
        character_ids = _sentences_to_ids(sentences)

        fast_out = fast(character_ids)
        allennlp_out = allennlp(character_ids)

        for repr_idx in range(2):
            fast_mixed_repr = fast_out['elmo_representations'][repr_idx]
            allennlp_mixed_repr = allennlp_out['elmo_representations'][repr_idx]
            np.testing.assert_array_almost_equal(
                    fast_mixed_repr,
                    allennlp_mixed_repr,
            )

        np.testing.assert_array_equal(
                fast_out['mask'],
                allennlp_out['mask'],
        )


def test_fast_elmo_with_allennlp_do_layer_norm():
    fast = FastElmo(
            ELMO_OPTIONS_FILE,
            ELMO_WEIGHT_FILE,
            num_output_representations=1,
            scalar_mix_parameters=[1.0, 1.0, 1.0],
            do_layer_norm=True,
    )

    allennlp = Elmo(
            ELMO_OPTIONS_FILE,
            ELMO_WEIGHT_FILE,
            num_output_representations=1,
            dropout=0.0,
            scalar_mix_parameters=[1.0, 1.0, 1.0],
            do_layer_norm=True,
    )

    sentences = [
            ['ELMo', 'helps', 'disambiguate', 'ELMo', 'from', 'Elmo', '.'],
            ['The', 'sentence', '.'],
    ]
    character_ids = _sentences_to_ids(sentences)

    fast_out = fast(character_ids)
    allennlp_out = allennlp(character_ids)

    # Since we don't include the BOS/EOS reprs during layer normalization,
    # the result will be different from AllenNLP's implementation.
    np.testing.assert_raises(
            AssertionError,
            np.testing.assert_array_almost_equal,
            fast_out['elmo_representations'][0],
            allennlp_out['elmo_representations'][0],
    )

    # We can pack BOS/EOS to inputs manually
    _beginning_of_sentence_characters = torch.from_numpy(
            np.array(ELMoCharacterMapper.beginning_of_sentence_characters) + 1)
    _end_of_sentence_characters = torch.from_numpy(
            np.array(ELMoCharacterMapper.end_of_sentence_characters) + 1)

    mask = ((character_ids > 0).long().sum(dim=-1) > 0).long()
    character_ids_with_bos_eos, mask_with_bos_eos = add_sentence_boundary_token_ids(
            character_ids,
            mask,
            _beginning_of_sentence_characters,
            _end_of_sentence_characters,
    )

    # And disable the mock BOS/EOS actions in FastElmo.
    fast.exec_managed_lstm_bos_eos = False
    fast_out_2 = fast(character_ids_with_bos_eos)
    fast_mixed_repr_2, _ = remove_sentence_boundaries(
            fast_out_2['elmo_representations'][0],
            fast_out_2['mask'],
    )

    allennlp_out_2 = allennlp(character_ids)

    np.testing.assert_array_almost_equal(
            fast_mixed_repr_2,
            allennlp_out_2['elmo_representations'][0],
    )


def test_fast_elmo_save_and_load():
    fast_1 = FastElmo(
            ELMO_OPTIONS_FILE,
            ELMO_WEIGHT_FILE,
    )

    # Change weight and save.
    fast_1.scalar_mix_0.scalar_parameters[0].data.fill_(42.)
    fast_1_state_dict = fast_1.state_dict()

    # Load.
    fast_2 = FastElmo(
            ELMO_OPTIONS_FILE,
            ELMO_WEIGHT_FILE,
    )
    fast_2.load_state_dict(fast_1_state_dict)

    assert float(fast_2.scalar_mix_0.scalar_parameters[0]) == 42.
    # assert float(fast_2.scalar_mixes[0].named_parameters()['scalar_0']) == 42.


def test_fast_elmo_word_embedding():
    vocab = utils.load_vocab(CACHE_VOCAB_FILE)

    fast_char_cnn = FastElmo(
            ELMO_OPTIONS_FILE,
            ELMO_WEIGHT_FILE,
    )

    fast_word_embd = FastElmoWordEmbedding(
            ELMO_OPTIONS_FILE,
            ELMO_WEIGHT_FILE,
            word_embedding_weight_file=CACHE_EMBD_FILE,
    )

    # Test BOS/EOS & other words.
    words = vocab

    embd_repr = fast_word_embd(utils.batch_to_word_ids([words], utils.build_vocab2id(vocab)))
    char_cnn_repr = fast_char_cnn(utils.batch_to_char_ids([words]))

    np.testing.assert_array_almost_equal(
            embd_repr['elmo_representations'][0].data.numpy(),
            char_cnn_repr['elmo_representations'][0].data.numpy(),
    )



from os.path import dirname, join

import torch
from allennlp.modules.elmo import batch_to_ids
import numpy as np

from pytorch_fast_elmo import utils
from pytorch_fast_elmo import FastElmoBase

FIXTURES_FODLER = join(dirname(__file__), 'fixtures')
ELMO_OPTIONS_FILE = join(FIXTURES_FODLER, 'options.json')
ELMO_WEIGHT_FILE = join(FIXTURES_FODLER, 'lm_weights.hdf5')


def test_batch_to_char_ids():
    sentences = [
            ["This", "is", "a", "sentence"],
            ["Here", "'s", "one"],
            ["Another", "one"],
    ]
    t1 = utils.batch_to_char_ids(sentences)
    t2 = batch_to_ids(sentences)
    np.testing.assert_array_equal(t1.numpy(), t2.numpy())

    sentences = [["one"]]
    t1 = utils.batch_to_char_ids(sentences)
    t2 = batch_to_ids(sentences)
    np.testing.assert_array_equal(t1.numpy(), t2.numpy())


def test_cache_char_cnn_vocab(tmpdir):
    vocab = ['<S>', '</S>', '<UNK>', 'ELMo', 'helps', 'disambiguate', 'ELMo', 'from', 'Elmo', '.']
    vocab_path = tmpdir.join("vocab.txt")
    vocab_path.write('\n'.join(vocab))

    embedding_path = tmpdir.join("embd.txt")

    utils.cache_char_cnn_vocab(
            vocab_path.realpath(),
            ELMO_OPTIONS_FILE,
            ELMO_WEIGHT_FILE,
            str(embedding_path.realpath()),
            batch_size=2,
            max_characters_per_token=15,
    )

    fast_word_embd = FastElmoBase(
            ELMO_OPTIONS_FILE,
            None,
            disable_word_embedding=False,
            word_embedding_weight_file=str(embedding_path.realpath()),
            # Disable all other components.
            disable_char_cnn=True,
            disable_forward_lstm=True,
            disable_backward_lstm=True,
            disable_scalar_mix=True,
    )
    fast_char_cnn = FastElmoBase(
            ELMO_OPTIONS_FILE,
            ELMO_WEIGHT_FILE,
            # Disable all other components.
            disable_forward_lstm=True,
            disable_backward_lstm=True,
            disable_scalar_mix=True,
    )

    embd_repr = fast_word_embd.exec_word_embedding(
            fast_word_embd.pack_inputs(
                    utils.batch_to_word_ids(
                            [['ELMo', 'helps', '!!!UNK!!!']],
                            utils.load_and_build_vocab2id(vocab_path.realpath()),
                    )))
    char_cnn_repr = fast_char_cnn.exec_char_cnn(
            fast_char_cnn.pack_inputs(
                    utils.batch_to_char_ids(
                            [['ELMo', 'helps', '<UNK>']],
                            max_characters_per_token=15,
                    )))

    np.testing.assert_array_almost_equal(
            embd_repr.data.numpy(),
            char_cnn_repr.data.numpy(),
    )
    np.testing.assert_array_equal(embd_repr.batch_sizes.numpy(), char_cnn_repr.batch_sizes.numpy())


def test_sort_batch_by_length():
    for _ in range(100):
        batch = torch.randn(20, 40)
        lengths = torch.randint(0, 41, (20,), dtype=torch.long)
        mask = utils.generate_mask_from_lengths(20, 40, lengths)
        batch[mask != 1] = 0

        sorted_batch, _, restoration_index = utils.sort_batch_by_length(batch)
        np.testing.assert_array_equal(
                sorted_batch.index_select(0, restoration_index),
                batch,
        )



from typing import Optional, Tuple, List, Dict, Any, Callable
import json
import logging
import itertools

import torch
import numpy as np
import h5py

from pytorch_fast_elmo import (
        batch_to_char_ids,
        load_and_build_vocab2id,
        batch_to_word_ids,
        FastElmoBase,
        FastElmo,
        FastElmoWordEmbedding,
        FastElmoPlainEncoder,
        FastElmoWordEmbeddingPlainEncoder,
        FastElmoForwardVocabDistrib,
        FastElmoBackwardVocabDistrib,
        FastElmoWordEmbeddingForwardVocabDistrib,
        FastElmoWordEmbeddingBackwardVocabDistrib,
)

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


def _generate_vocab2id_id2vocab(vocab_txt: str,) -> Tuple[Dict[str, int], Dict[int, str]]:
    vocab2id = load_and_build_vocab2id(vocab_txt)
    id2vocab = {token_id: token for token, token_id in vocab2id.items()}
    return vocab2id, id2vocab


def _generate_batch_to_ids(
        vocab2id: Dict[str, int],
        char_cnn_maxlen: int,
        no_char_cnn: bool,
        cuda_device: int,
) -> Callable[[List[List[str]]], torch.Tensor]:
    if no_char_cnn:

        def batch_to_ids(batch: List[List[str]]) -> torch.Tensor:
            tensor = batch_to_word_ids(batch, vocab2id)
            if cuda_device >= 0:
                tensor = tensor.cuda(cuda_device)
            return tensor
    else:

        def batch_to_ids(batch: List[List[str]]) -> torch.Tensor:
            if char_cnn_maxlen == 0:
                tensor = batch_to_char_ids(batch)
            else:
                tensor = batch_to_char_ids(batch, char_cnn_maxlen)
            if cuda_device >= 0:
                tensor = tensor.cuda(cuda_device)
            return tensor

    return batch_to_ids


def _warm_up(
        warm_up_txt: str,
        batch_to_ids: Callable[[List[List[str]]], torch.Tensor],
        elmo: FastElmoBase,
) -> None:
    sentences_token_ids = []
    with open(warm_up_txt) as fin:
        for line in fin:
            sent = line.split()
            if not sent:
                continue
            token_ids = batch_to_ids([sent])
            sentences_token_ids.append(token_ids)

    for token_ids in sentences_token_ids:
        with torch.no_grad():
            elmo(token_ids)


def sample_sentence(
        options_file: str,
        weight_file: str,
        vocab_txt: str,
        output_json: str,
        enable_trace: bool,
        no_char_cnn: bool,
        char_cnn_maxlen: int,
        go_forward: bool,
        next_token_top_k: int,
        sample_size: int,
        sample_constrain_txt: Optional[str],
        warm_up_txt: Optional[str],
        cuda_device: int,
) -> None:
    if no_char_cnn:
        if go_forward:
            fast_elmo_cls = FastElmoWordEmbeddingForwardVocabDistrib
        else:
            fast_elmo_cls = FastElmoWordEmbeddingBackwardVocabDistrib
    else:
        if go_forward:
            fast_elmo_cls = FastElmoForwardVocabDistrib
        else:
            fast_elmo_cls = FastElmoBackwardVocabDistrib

    vocab2id, id2vocab = _generate_vocab2id_id2vocab(vocab_txt)
    batch_to_ids = _generate_batch_to_ids(
            vocab2id,
            char_cnn_maxlen,
            no_char_cnn,
            cuda_device,
    )

    elmo = fast_elmo_cls(options_file, weight_file)
    if cuda_device >= 0:
        elmo = elmo.cuda(cuda_device)

    # Warm up.
    if warm_up_txt:
        _warm_up(warm_up_txt, batch_to_ids, elmo)

    # Manually deal with BOS/EOS.
    elmo.exec_managed_lstm_bos_eos = False

    if sample_constrain_txt:
        with open(sample_constrain_txt) as fin:
            lines = fin.readlines()
            if not lines:
                raise ValueError('No content in sample_constrain_txt.')
            if len(lines) > 1:
                logging.warning('Multiple lines in sample_constrain_txt, only use the 1st line.')
            sample_constrain_tokens = lines[0].split()
        if not go_forward:
            sample_constrain_tokens.reverse()

    infos: List[Any] = []
    for _ in range(sample_size):
        if go_forward:
            cur_token, end_token = '<S>', '</S>'
        else:
            cur_token, end_token = '</S>', '<S>'

        if sample_constrain_txt:
            for token in itertools.chain([cur_token], sample_constrain_tokens[:-1]):
                with torch.no_grad():
                    elmo(batch_to_ids([[token]]))
            cur_token = sample_constrain_tokens[-1]

        info: List[Any] = []
        while cur_token != end_token:
            batched = batch_to_ids([[cur_token]])
            with torch.no_grad():
                output, _ = elmo(batched)
            if cuda_device >= 0:
                output = output.cpu()

            probs, indices = torch.topk(output.view(-1), next_token_top_k)
            probs = probs.numpy()
            indices = indices.numpy()

            next_token_id = np.random.choice(indices, p=probs / probs.sum()) + 1
            next_token = id2vocab[next_token_id]

            info_probs = sorted(
                    dict(zip(map(id2vocab.get, indices + 1), probs.tolist())).items(),
                    key=lambda p: p[1],
                    reverse=True,
            )
            info.append({
                    'cur': cur_token,
                    'next': next_token,
                    'probs': info_probs,
            })

            cur_token = next_token

        # Ending.
        with torch.no_grad():
            elmo(batch_to_ids([[end_token]]))
        # Save info.
        infos.append({'text': ''.join(step['cur'] for step in info)})
        if enable_trace:
            infos[-1]['trace'] = info
        if sample_constrain_txt:
            infos[-1]['text'] = ''.join(sample_constrain_tokens[:-1]) + infos[-1]['text']
        if not go_forward:
            infos[-1]['text'] = infos[-1]['text'][::-1]

    # Output to JSON.
    with open(output_json, 'w') as fout:
        json.dump(infos, fout, ensure_ascii=False, indent=2)


def encode_sentences(
        options_file: str,
        weight_file: str,
        vocab_txt: str,
        input_txt: str,
        output_hdf5: str,
        no_char_cnn: bool,
        char_cnn_maxlen: int,
        scalar_mix: Optional[Tuple[float]],
        warm_up_txt: Optional[str],
        cuda_device: int,
) -> None:
    if scalar_mix is None:
        if no_char_cnn:
            fast_elmo_cls = FastElmoWordEmbeddingPlainEncoder
        else:
            fast_elmo_cls = FastElmoPlainEncoder

        elmo = fast_elmo_cls(
                options_file,
                weight_file,
        )

    else:
        if no_char_cnn:
            fast_elmo_cls = FastElmoWordEmbedding
        else:
            fast_elmo_cls = FastElmo

        elmo = fast_elmo_cls(
                options_file,
                weight_file,
                scalar_mix_parameters=list(scalar_mix),
        )

    if cuda_device >= 0:
        elmo = elmo.cuda(cuda_device)

    vocab2id, _ = _generate_vocab2id_id2vocab(vocab_txt)
    batch_to_ids = _generate_batch_to_ids(
            vocab2id,
            char_cnn_maxlen,
            no_char_cnn,
            cuda_device,
    )

    # Warm up.
    if warm_up_txt:
        _warm_up(warm_up_txt, batch_to_ids, elmo)

    sentences: List[Tuple[int, List[str]]] = []
    with open(input_txt) as fin:
        for sentence_id, line in enumerate(fin):
            tokens = line.split()
            if not tokens:
                logger.warning('Ignore sentence_id = %s', sentence_id)
                continue
            sentences.append((sentence_id, tokens))

    with h5py.File(output_hdf5, 'w') as fout:
        for sentence_id, tokens in sentences:
            token_ids = batch_to_ids([tokens])

            if scalar_mix is None:
                with torch.no_grad():
                    layer_reprs, _ = elmo(token_ids)
                # (layers, timesteps, hidden_size)
                encoded = torch.cat(layer_reprs, dim=0)
            else:
                with torch.no_grad():
                    out = elmo(token_ids)
                # (1, timesteps, hidden_size)
                encoded = out['elmo_representations'][0]

            if cuda_device >= 0:
                encoded = encoded.cpu()

            fout.create_dataset(
                    str(sentence_id),
                    encoded.shape,
                    dtype='float32',
                    data=encoded.numpy(),
            )



from typing import List, Tuple, Iterable, Dict, Optional

import torch
from torch.nn.utils.rnn import pack_padded_sequence, pad_packed_sequence, PackedSequence
import numpy as np

from pytorch_fast_elmo.factory import ElmoCharacterEncoderFactory, ElmoWordEmbeddingFactory
from _pytorch_fast_elmo import ElmoCharacterEncoder  # pylint: disable=no-name-in-module


def load_vocab(vocab_txt: str) -> List[str]:
    """
    Use the same format as bilm-tf.
    """
    vocab = []
    with open(vocab_txt) as fin:
        for line in fin:
            word = line.strip()
            if word:
                vocab.append(word)
    return vocab


def build_vocab2id(vocab: List[str]) -> Dict[str, int]:
    """
    Adding one will be applied for padding.
    """
    assert len(vocab) > 3
    assert vocab[:3] == ['<S>', '</S>', '<UNK>']
    return {word: word_id for word_id, word in enumerate(vocab, start=1)}


def load_and_build_vocab2id(vocab_txt: str) -> Dict[str, int]:
    return build_vocab2id(load_vocab(vocab_txt))


def get_lengths_of_zero_padded_batch(inputs: torch.Tensor) -> torch.Tensor:
    if inputs.dim() == 2:
        lengths = (inputs > 0).long().sum(dim=-1)
    elif inputs.dim() == 3:
        lengths = ((inputs > 0).long().sum(dim=-1) > 0).long().sum(dim=-1)
    else:
        raise ValueError("inputs should be 2D or 3D.")

    return lengths


def pack_inputs(
        inputs: torch.Tensor,
        lengths: Optional[torch.Tensor] = None,
) -> PackedSequence:
    """
    Pack inputs of shape `(batch_size, timesteps, x)` or `(batch_size, timesteps)`.
    Padding value should be 0.
    """
    if lengths is None:
        lengths = get_lengths_of_zero_padded_batch(inputs)
    return pack_padded_sequence(inputs, lengths, batch_first=True)


def generate_mask_from_lengths(
        batch_size: int,
        max_timesteps: int,
        lengths: torch.Tensor,
) -> torch.Tensor:
    ones = lengths.new_ones(batch_size, max_timesteps, dtype=torch.long)
    range_tensor = ones.cumsum(dim=-1)
    return (lengths.unsqueeze(1) >= range_tensor).long()


def unpack_outputs(inputs: PackedSequence) -> torch.Tensor:
    """
    Unpack the final result and return `(tensor, mask)`.
    """
    tensor, _ = pad_packed_sequence(inputs, batch_first=True)
    return tensor


class ElmoCharacterIdsConst:
    """
    From Allennlp.
    """
    MAX_WORD_LENGTH = 50

    BEGINNING_OF_SENTENCE_CHARACTER = 256  # <begin sentence>
    END_OF_SENTENCE_CHARACTER = 257  # <end sentence>
    BEGINNING_OF_WORD_CHARACTER = 258  # <begin word>
    END_OF_WORD_CHARACTER = 259  # <end word>
    PADDING_CHARACTER = 260  # <padding>


def make_padded_char_ids(
        char_ids: Iterable[int],  # +1 should have been applied.
        max_word_length: int = ElmoCharacterIdsConst.MAX_WORD_LENGTH,
        padding_character: int = ElmoCharacterIdsConst.PADDING_CHARACTER + 1,
        beginning_of_word_character: int = ElmoCharacterIdsConst.BEGINNING_OF_WORD_CHARACTER + 1,
        end_of_word_character: int = ElmoCharacterIdsConst.END_OF_WORD_CHARACTER + 1,
) -> List[int]:
    padded = [padding_character] * max_word_length

    padded[0] = beginning_of_word_character
    idx = 1
    for char_id in char_ids:
        if idx >= max_word_length:
            break
        padded[idx] = char_id
        idx += 1

    idx = min(idx, max_word_length - 1)
    padded[idx] = end_of_word_character

    return padded


def make_bos(max_word_length: int = ElmoCharacterIdsConst.MAX_WORD_LENGTH) -> List[int]:
    return make_padded_char_ids(
            (ElmoCharacterIdsConst.BEGINNING_OF_SENTENCE_CHARACTER + 1,),
            max_word_length,
    )


def make_eos(max_word_length: int = ElmoCharacterIdsConst.MAX_WORD_LENGTH) -> List[int]:
    return make_padded_char_ids(
            (ElmoCharacterIdsConst.END_OF_SENTENCE_CHARACTER + 1,),
            max_word_length,
    )


def word_to_char_ids(word: str) -> List[int]:
    # +1 is applied here.
    return [char_id + 1 for char_id in word.encode('utf-8', 'ignore')]


_WORD_TO_CHAR_IDS_EXCEPTION = {
        '<S>': make_bos,
        '</S>': make_eos,
}


def batch_to_char_ids(
        batch: List[List[str]],
        max_characters_per_token: int = ElmoCharacterIdsConst.MAX_WORD_LENGTH,
) -> torch.Tensor:
    """
    From Allennlp.

    Note:
    1. BOS/EOS will be treated specially.
    2. UNK will be treated as normal string, same as bilm-tf.

    Return tensor of shape `(batch_size, max_timesteps, max_characters_per_token)`.
    """
    max_timesteps = max(len(row) for row in batch)
    zeros = torch.LongTensor([0] * max_characters_per_token)

    rows = []
    for words in batch:
        row = []
        for word in words:
            special_gen = _WORD_TO_CHAR_IDS_EXCEPTION.get(word)
            if special_gen is None:
                char_ids = make_padded_char_ids(
                        word_to_char_ids(word),
                        max_characters_per_token,
                )
            else:
                char_ids = special_gen(max_characters_per_token)

            # of shape `(max_characters_per_token,)`
            row.append(torch.LongTensor(char_ids))

        # Add padding.
        row.extend([zeros] * (max_timesteps - len(row)))
        # Stack to shape `(max_timesteps, max_characters_per_token)`
        rows.append(torch.stack(row))

    # Stack to shape `(batch_size, max_timesteps, max_characters_per_token)`
    return torch.stack(rows)


def batch_to_word_ids(batch: List[List[str]], vocab2id: Dict[str, int]) -> torch.Tensor:
    """
    For word embedding.

    1. UNK will be mapped to 3 since we assume the vocab starts with `<S>, </S>, <UNK>`.

    Return tensor of shape `(batch_size, max_timesteps)`.
    """
    max_timesteps = max(len(row) for row in batch)

    rows = []
    for words in batch:
        row = [vocab2id.get(word, 3) for word in words]
        row.extend([0] * (max_timesteps - len(row)))
        rows.append(row)

    return torch.LongTensor(rows)


def get_bos_eos_token_repr(
        char_cnn_factory: ElmoCharacterEncoderFactory,
        char_cnn: ElmoCharacterEncoder,
) -> Tuple[torch.Tensor, torch.Tensor]:
    # [<bow>, <bos/eos>, <eow>, max(kernal)...]
    max_characters_per_token = max(kernal_size for kernal_size, _ in char_cnn_factory.filters)
    max_characters_per_token += 3

    bos_ids = make_bos(max_characters_per_token)
    eos_ids = make_eos(max_characters_per_token)
    bos_eos = torch.LongTensor([bos_ids, eos_ids])

    char_cnn_param = char_cnn.parameters()[0]
    if char_cnn_param.is_cuda:
        bos_eos = bos_eos.cuda(char_cnn_param.get_device())

    with torch.no_grad():
        bos_eos_reprs = char_cnn(bos_eos)

    return bos_eos_reprs[0], bos_eos_reprs[1]


def sort_batch_by_length(
        batch: torch.Tensor,
        lengths: Optional[torch.Tensor] = None,
) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
    """
    Similar to AllenNLP.

    `batch` of shape `(batch_size, max_timesteps, *)` should be zero padded.
    `sequence_lengths` of shape `(batch_size,)`

    Returns (sorted_batch, restoration_index)
    """
    if lengths is None:
        lengths = get_lengths_of_zero_padded_batch(batch)
    _, permutation_index = lengths.sort(0, descending=True)
    sorted_batch = batch.index_select(0, permutation_index)
    _, restoration_index = permutation_index.sort(0, descending=False)
    return sorted_batch, permutation_index, restoration_index


def export_word_embedding_to_txt(
        vocab: List[str],
        embedding_weight: np.array,
        txt_out: str,
) -> None:
    if embedding_weight.shape[0] != len(vocab):
        raise ValueError('Size not match. '
                         f'embd({embedding_weight.shape[0]}) '
                         '!= '
                         f'vocab({len(vocab)})')

    with open(txt_out, 'w') as fout:
        for idx, token in enumerate(vocab):
            embd = embedding_weight[idx]
            # [token] [dim1] [dim2]...
            embd_txt = ' '.join(map(str, embd))
            line = f'{token} {embd_txt}\n'
            fout.write(line)


def cache_char_cnn_vocab(
        vocab_txt: str,
        options_file: str,
        weight_file: str,
        txt_out: str,
        max_characters_per_token: int = ElmoCharacterIdsConst.MAX_WORD_LENGTH,
        cuda_device: int = -1,
        batch_size: int = 256,
) -> None:
    """
    1. Load vocab.
    2. Feed vocab to Char CNN.
    3. Feed BOS/EOS to Char CNN.
    4. Dump reprs to txt. (will be loaded by `ElmoWordEmbeddingFactory`).
    """
    # 1.
    vocab = load_vocab(vocab_txt)

    # 2.
    char_cnn_factory = ElmoCharacterEncoderFactory(
            options_file,
            weight_file,
    )
    char_cnn = char_cnn_factory.create(requires_grad=False)
    if cuda_device >= 0:
        char_cnn.cuda(cuda_device)

    cached = []
    for batch_start in range(0, len(vocab), batch_size):
        batch = vocab[batch_start:batch_start + batch_size]
        # (1, batch_size, max_characters_per_token)
        char_ids = batch_to_char_ids([batch], max_characters_per_token)
        if cuda_device >= 0:
            char_ids = char_ids.cuda(cuda_device)

        inputs = pack_inputs(char_ids)
        output_data = char_cnn(inputs.data)
        # (1, batch_size, output_dim)
        char_reprs = unpack_outputs(PackedSequence(output_data, inputs.batch_sizes))
        # (batch_size, output_dim)
        cached.append(char_reprs.squeeze(0))

    # (vocab, output_dim)
    combined = torch.cat(cached, dim=0)
    if cuda_device >= 0:
        combined = combined.cpu()
    embedding_weight = combined.numpy()

    # 3.
    lstm_bos_repr, lstm_eos_repr = get_bos_eos_token_repr(
            char_cnn_factory,
            char_cnn,
    )
    if cuda_device >= 0:
        lstm_bos_repr = lstm_bos_repr.cpu()
        lstm_eos_repr = lstm_eos_repr.cpu()

    lstm_bos_weight = lstm_bos_repr.numpy()
    lstm_eos_weight = lstm_eos_repr.numpy()

    embedding_weight[0] = lstm_bos_weight
    embedding_weight[1] = lstm_eos_weight

    # 4.
    export_word_embedding_to_txt(vocab, embedding_weight, txt_out)


def export_word_embd(
        vocab_txt: str,
        weight_file: str,
        txt_out: str,
) -> None:
    vocab = load_vocab(vocab_txt)
    word_embedding_factory = ElmoWordEmbeddingFactory(None, weight_file)
    embedding_weight, _, _ = word_embedding_factory.create(requires_grad=False)
    # remove padding.
    embedding_weight = embedding_weight.data[1:, :].numpy()

    export_word_embedding_to_txt(vocab, embedding_weight, txt_out)



"""
Follows AllenNLP.
"""
# pylint: disable=attribute-defined-outside-init

from typing import Dict, Tuple, Any, Optional, List
import json
import math

import torch
import h5py
import numpy as np

from pytorch_stateful_lstm import StatefulUnidirectionalLstm
from _pytorch_fast_elmo import ElmoCharacterEncoder  # pylint: disable=no-name-in-module


def load_options(options_file: Optional[str]):  # type: ignore
    if options_file is None:
        return None
    else:
        with open(options_file) as fin:
            return json.load(fin)


def freeze_parameters(named_parameters: Dict[str, torch.Tensor]) -> None:
    for param in named_parameters.values():
        param.requires_grad = False


class FactoryBase:

    def __init__(
            self,
            options_file: Optional[str],
            weight_file: Optional[str],
    ) -> None:
        self.options = load_options(options_file)
        self.weight_file = weight_file


class ElmoCharacterEncoderFactory(FactoryBase):

    @staticmethod
    def from_scratch(
            char_embedding_cnt: int,
            char_embedding_dim: int,
            filters: List[Tuple[int, int]],
            activation: str,
            num_highway_layers: int,
            output_dim: int,
    ) -> 'ElmoCharacterEncoderFactory':
        factory = ElmoCharacterEncoderFactory(None, None)
        factory.options = {
                'n_characters': char_embedding_cnt,
                'char_cnn': {
                        'embedding': {
                                'dim': char_embedding_dim
                        },
                        'filters': filters,
                        'activation': activation,
                        'n_highway': num_highway_layers,
                },
                'lstm': {
                        'projection_dim': output_dim
                },
        }
        return factory

    def create(self, requires_grad: bool = False) -> ElmoCharacterEncoder:
        assert self.options and 'char_cnn' in self.options

        # Collect parameters for the construction of `ElmoCharacterEncoder`.
        self.char_embedding_cnt = self.options.get('n_characters', 261)
        self.char_embedding_dim = self.options['char_cnn']['embedding']['dim']
        self.filters = self.options['char_cnn']['filters']
        self.activation = self.options['char_cnn']['activation']
        self.num_highway_layers = self.options['char_cnn']['n_highway']
        self.output_dim = self.options['lstm']['projection_dim']

        self.named_parameters: Dict[str, torch.Tensor] = {}

        module = ElmoCharacterEncoder(
                self.char_embedding_cnt,
                self.char_embedding_dim,
                self.filters,
                self.activation,
                self.num_highway_layers,
                self.output_dim,
        )
        self.named_parameters.update(module.named_parameters())

        if self.weight_file:
            self._load_char_embedding()
            self._load_cnn_weights()
            self._load_highway()
            self._load_projection()
        else:
            assert requires_grad

        if not requires_grad:
            freeze_parameters(self.named_parameters)

        return module

    def _load_char_embedding(self) -> None:
        with h5py.File(self.weight_file, 'r') as fin:
            char_embed_weights = fin['char_embed'][...]

        if char_embed_weights.shape != \
                (self.char_embedding_cnt, self.char_embedding_dim):
            raise ValueError('Char embd shape not match. '
                             f'Loaded shape: {char_embed_weights.shape}')

        self.named_parameters['char_embedding.weight'].data[0:, :] = \
                torch.zeros(1, self.char_embedding_dim, dtype=torch.float)
        self.named_parameters['char_embedding.weight'].data[1:, :] = \
                torch.FloatTensor(char_embed_weights)

    def _load_cnn_weights(self) -> None:
        for conv_idx, (kernel_size, out_channels) in enumerate(self.filters):
            with h5py.File(self.weight_file, 'r') as fin:
                weight = fin['CNN'][f'W_cnn_{conv_idx}'][...]
                bias = fin['CNN'][f'b_cnn_{conv_idx}'][...]

            w_reshaped = np.transpose(weight.squeeze(axis=0), axes=(2, 1, 0))
            if w_reshaped.shape != (out_channels, self.char_embedding_dim, kernel_size):
                raise ValueError("Invalid weight file")

            weight_name = f'char_conv_{conv_idx}.weight'
            self.named_parameters[weight_name].data.copy_(torch.FloatTensor(w_reshaped))

            if bias.shape != (out_channels,):
                raise ValueError("Invalid weight file")

            bias_name = f'char_conv_{conv_idx}.bias'
            self.named_parameters[bias_name].data.copy_(torch.FloatTensor(bias))

    def _load_highway(self) -> None:
        """
        Note: `W_carry` and `b_carry` in bilm-tf are weights of transform gate.
        """
        total_out_channels = sum(out_channels for _, out_channels in self.filters)

        for layer_idx in range(self.num_highway_layers):
            with h5py.File(self.weight_file, 'r') as fin:
                w_transform = np.transpose(fin[f'CNN_high_{layer_idx}']['W_transform'][...])
                w_transform_gate = np.transpose(fin[f'CNN_high_{layer_idx}']['W_carry'][...])
                weight = np.concatenate([w_transform, w_transform_gate], axis=0)

                if weight.shape != (total_out_channels * 2, total_out_channels):
                    raise ValueError("Invalid weight file")

                weight_name = f'highway.layers_{layer_idx}.weight'
                self.named_parameters[weight_name].data.copy_(torch.FloatTensor(weight))

                b_transform = fin[f'CNN_high_{layer_idx}']['b_transform'][...]
                b_transform_gate = fin[f'CNN_high_{layer_idx}']['b_carry'][...]
                bias = np.concatenate([b_transform, b_transform_gate], axis=0)

                if bias.shape != (total_out_channels * 2,):
                    raise ValueError("Invalid weight file")

                bias_name = f'highway.layers_{layer_idx}.bias'
                self.named_parameters[bias_name].data.copy_(torch.FloatTensor(bias))

    def _load_projection(self) -> None:
        total_out_channels = sum(out_channels for _, out_channels in self.filters)

        with h5py.File(self.weight_file, 'r') as fin:
            weight = fin['CNN_proj']['W_proj'][...]
            bias = fin['CNN_proj']['b_proj'][...]

        weight = np.transpose(weight)
        if weight.shape != (self.output_dim, total_out_channels):
            raise ValueError("Invalid weight file")

        weight_name = 'output_proj.weight'
        self.named_parameters[weight_name].data.copy_(torch.FloatTensor(weight))

        if bias.shape != (self.output_dim,):
            raise ValueError("Invalid weight file")

        bias_name = f'output_proj.bias'
        self.named_parameters[bias_name].data.copy_(torch.FloatTensor(bias))


class ElmoWordEmbeddingFactory(FactoryBase):

    @staticmethod
    def from_scratch(
            cnt: int,
            dim: int,
    ) -> 'ElmoWordEmbeddingFactory':
        factory = ElmoWordEmbeddingFactory(None, None)
        factory.options = {
                'n_tokens_vocab': cnt,
                'word_embedding_dim': dim,
        }
        return factory

    def weight_file_is_hdf5(self) -> bool:
        try:
            with h5py.File(self.weight_file, 'r') as fin:
                return True
        except OSError as ex:
            return False

    def create(
            self,
            requires_grad: bool = False,
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        """
        Returns (embedding, lstm_bos, lstm_eos)
        """
        if self.weight_file:
            # Load `embd_weight` from hdf5 or txt.
            if self.weight_file_is_hdf5():
                # HDF5 format.
                with h5py.File(self.weight_file, 'r') as fin:
                    assert 'embedding' in fin
                    embd_weight = fin['embedding'][...]

            else:
                # TXT format.
                loaded_cnt = 0
                loaded_dim = 0
                loaded_embds = []
                with open(self.weight_file) as fin:
                    for idx, line in enumerate(fin):
                        fields = line.split()
                        if not fields:
                            continue

                        # L0: <cnt> <dim>
                        if idx == 0 and len(fields) == 2:
                            loaded_dim = int(fields[1])
                            continue

                        token = fields[0]
                        embd = fields[1:]
                        loaded_cnt += 1

                        if loaded_dim == 0:
                            loaded_dim = len(embd)
                        elif loaded_dim != len(embd):
                            raise ValueError(f'Dimension not match on L{idx}: {token}, '
                                             f'should be {loaded_dim}-D.')

                        vec = np.expand_dims(
                                np.asarray(list(map(float, embd))),
                                0,
                        )
                        loaded_embds.append(vec)

                embd_weight = np.concatenate(loaded_embds)

            # Since bilm-tf doesn't include padding,
            # we need to prepend a padding row in index 0.
            self.word_embedding_cnt = embd_weight.shape[0]
            self.word_embedding_dim = embd_weight.shape[1]

            # Check with options if `n_tokens_vocab` exists.
            if 'n_tokens_vocab' in self.options \
                    and self.options['n_tokens_vocab'] != self.word_embedding_cnt:
                raise ValueError('n_tokens_vocab not match')

            embd = torch.zeros(
                    (self.word_embedding_cnt + 1, self.word_embedding_dim),
                    dtype=torch.float,
            )

            embd.data[1:, :].copy_(torch.FloatTensor(embd_weight))
            embd.requires_grad = requires_grad

            lstm_bos_repr = embd.data[1]
            lstm_eos_repr = embd.data[2]

        else:
            assert requires_grad
            assert self.options['n_tokens_vocab'] > 0

            self.word_embedding_cnt = self.options['n_tokens_vocab']
            self.word_embedding_dim = self.options['word_embedding_dim']

            embd = torch.zeros(
                    (self.word_embedding_cnt + 1, self.word_embedding_dim),
                    dtype=torch.float,
            )
            torch.nn.init.normal_(embd)
            embd.requires_grad = True

            # `exec_managed_lstm_bos_eos` should be disabled in this case.
            lstm_bos_repr = None
            lstm_eos_repr = None

        return embd, lstm_bos_repr, lstm_eos_repr


class ElmoLstmFactory(FactoryBase):

    @staticmethod
    def from_scratch(
            num_layers: int,
            input_size: int,
            hidden_size: int,
            cell_size: int,
            cell_clip: float,
            proj_clip: float,
            truncated_bptt: int,
    ) -> 'ElmoLstmFactory':
        factory = ElmoLstmFactory(None, None)
        factory.options = {
                'lstm': {
                        'n_layers': num_layers,
                        'projection_dim': input_size,
                        '_hidden_size': hidden_size,
                        'dim': cell_size,
                        'cell_clip': cell_clip,
                        'proj_clip': proj_clip,
                },
                'unroll_steps': truncated_bptt,
        }
        return factory

    def create(
            self,
            enable_forward: bool = False,
            forward_requires_grad: bool = False,
            enable_backward: bool = False,
            backward_requires_grad: bool = False,
    ) -> Tuple[StatefulUnidirectionalLstm, StatefulUnidirectionalLstm]:
        assert self.options and 'lstm' in self.options

        self.num_layers = self.options['lstm']['n_layers']
        self.input_size = self.options['lstm']['projection_dim']
        self.cell_size = self.options['lstm']['dim']
        self.cell_clip = self.options['lstm']['cell_clip']
        self.proj_clip = self.options['lstm']['proj_clip']
        self.truncated_bptt = self.options.get('unroll_steps', 20)
        self.use_skip_connections = True

        if self.options['lstm'].get('_hidden_size', 0) > 0:
            self.hidden_size = self.options['lstm']['_hidden_size']
        else:
            self.hidden_size = self.input_size

        self.named_parameters: Dict[str, torch.Tensor] = {}

        fwd_lstm = None
        if enable_forward:
            fwd_lstm = StatefulUnidirectionalLstm(
                    go_forward=True,
                    num_layers=self.num_layers,
                    input_size=self.input_size,
                    hidden_size=self.hidden_size,
                    cell_size=self.cell_size,
                    cell_clip=self.cell_clip,
                    proj_clip=self.proj_clip,
                    truncated_bptt=self.truncated_bptt,
                    use_skip_connections=self.use_skip_connections,
            )
            fwd_lstm_named_parameters = fwd_lstm.named_parameters()
            self.named_parameters.update(fwd_lstm_named_parameters)

        bwd_lstm = None
        if enable_backward:
            bwd_lstm = StatefulUnidirectionalLstm(
                    go_forward=False,
                    num_layers=self.num_layers,
                    input_size=self.input_size,
                    hidden_size=self.hidden_size,
                    cell_size=self.cell_size,
                    cell_clip=self.cell_clip,
                    proj_clip=self.proj_clip,
                    truncated_bptt=self.truncated_bptt,
                    use_skip_connections=self.use_skip_connections,
            )
            bwd_lstm_named_parameters = bwd_lstm.named_parameters()
            self.named_parameters.update(bwd_lstm_named_parameters)

        if enable_forward and enable_backward:
            if set(fwd_lstm_named_parameters.keys()) & \
                    set(bwd_lstm_named_parameters.keys()):
                raise ValueError('key conflict.')

        # Load weights.
        if self.weight_file:
            with h5py.File(self.weight_file, 'r') as fin:
                for layer_idx in range(self.num_layers):
                    for direction, prefix in enumerate([
                            'uni_lstm.forward_layer_',
                            'uni_lstm.backward_layer_',
                    ]):
                        good_forward = (direction == 0 and enable_forward)
                        good_backward = (direction == 1 and enable_backward)
                        if good_forward or good_backward:
                            dataset = fin[f'RNN_{direction}']\
                                        ['RNN']\
                                        ['MultiRNNCell']\
                                        [f'Cell{layer_idx}']\
                                        ['LSTMCell']
                            self._load_lstm(prefix + str(layer_idx), dataset)
        else:
            if enable_forward:
                assert forward_requires_grad
            if enable_backward:
                assert backward_requires_grad

        if enable_forward and not forward_requires_grad:
            freeze_parameters(fwd_lstm_named_parameters)

        if enable_backward and not backward_requires_grad:
            freeze_parameters(bwd_lstm_named_parameters)

        return fwd_lstm, bwd_lstm

    def _load_lstm(self, prefix: str, dataset: Any) -> None:
        cell_size = self.cell_size
        input_size = self.input_size

        tf_weights = np.transpose(dataset['W_0'][...])
        torch_weights = tf_weights.copy()

        input_weights = torch_weights[:, :input_size]
        recurrent_weights = torch_weights[:, input_size:]
        tf_input_weights = tf_weights[:, :input_size]
        tf_recurrent_weights = tf_weights[:, input_size:]

        for torch_w, tf_w in [[input_weights, tf_input_weights],
                              [recurrent_weights, tf_recurrent_weights]]:
            torch_w[(1 * cell_size):(2 * cell_size), :] = tf_w[(2 * cell_size):(3 * cell_size), :]
            torch_w[(2 * cell_size):(3 * cell_size), :] = tf_w[(1 * cell_size):(2 * cell_size), :]

        self.named_parameters[prefix + '.input_linearity_weight'].data.copy_(
                torch.FloatTensor(input_weights),)
        self.named_parameters[prefix + '.hidden_linearity_weight'].data.copy_(
                torch.FloatTensor(recurrent_weights),)

        tf_bias = dataset['B'][...]
        tf_bias[(2 * cell_size):(3 * cell_size)] += 1
        torch_bias = tf_bias.copy()
        torch_bias[(1 * cell_size):(2 * cell_size)] = tf_bias[(2 * cell_size):(3 * cell_size)]
        torch_bias[(2 * cell_size):(3 * cell_size)] = tf_bias[(1 * cell_size):(2 * cell_size)]

        self.named_parameters[prefix + '.hidden_linearity_bias'].data.copy_(
                torch.FloatTensor(torch_bias),)

        proj_weights = np.transpose(dataset['W_P_0'][...])

        self.named_parameters[prefix + '.proj_linearity_weight'].data.copy_(
                torch.FloatTensor(proj_weights),)


class ElmoVocabProjectionFactory(FactoryBase):

    @staticmethod
    def from_scratch(
            input_size: int,
            proj_size: int,
    ) -> 'ElmoVocabProjectionFactory':
        factory = ElmoVocabProjectionFactory(None, None)
        factory.options = {
                'lstm': {
                        'projection_dim': input_size
                },
                'n_tokens_vocab': proj_size,
        }
        return factory

    def create(
            self,
            requires_grad: bool = False,
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Returns (weight, bias) for affine transformation.
        """
        assert self.options \
                and 'n_tokens_vocab' in self.options \
                and 'lstm' in self.options

        self.input_size = self.options['lstm']['projection_dim']
        self.proj_size = self.options['n_tokens_vocab']
        assert self.input_size > 0 and self.proj_size > 0

        # Note: no padding zero.
        weight = torch.zeros(
                (self.proj_size, self.input_size),
                dtype=torch.float,
        )
        bias = torch.zeros(
                (self.proj_size,),
                dtype=torch.float,
        )

        if self.weight_file:
            with h5py.File(self.weight_file, 'r') as fin:
                if 'softmax' not in fin:
                    raise ValueError('softmax not in weight file.')
                loaded_weight = fin['softmax']['W'][...]
                loaded_bias = fin['softmax']['b'][...]

            weight.data.copy_(torch.FloatTensor(loaded_weight))
            weight.requires_grad = requires_grad

            bias.data.copy_(torch.FloatTensor(loaded_bias))
            bias.requires_grad = requires_grad

        else:
            assert requires_grad
            # init.
            torch.nn.init.kaiming_uniform_(weight, a=math.sqrt(5))
            weight.requires_grad = True

            fan_in, _ = torch.nn.init._calculate_fan_in_and_fan_out(weight)  # pylint: disable=protected-access
            bound = 1 / math.sqrt(fan_in)
            torch.nn.init.uniform_(bias, -bound, bound)
            bias.requires_grad = True

        return weight, bias



"""
Provide helper classes/functions to execute ELMo.
"""
# pylint: disable=no-self-use,arguments-differ,too-many-public-methods,too-many-lines
from typing import List, Tuple, Optional, Dict, Union, Any, Set
from collections import OrderedDict

import torch
from torch.nn.utils.rnn import PackedSequence
from torch.nn import ParameterList, Parameter

from pytorch_fast_elmo.factory import (
        ElmoCharacterEncoderFactory,
        ElmoWordEmbeddingFactory,
        ElmoLstmFactory,
        ElmoVocabProjectionFactory,
)
from pytorch_fast_elmo import utils

# TODO: use in inference.
# from _pytorch_fast_elmo import ScalarMix  # pylint: disable=no-name-in-module


def _raise_if_kwargs_is_invalid(allowed: Set[str], kwargs: Dict[str, Any]) -> None:
    invalid_keys = set(kwargs) - allowed
    if invalid_keys:
        msg = '\n'.join('invalid kwargs: {}'.format(key) for key in invalid_keys)
        raise ValueError(msg)


# Implement in Python as a temporary solution.
class ScalarMix(torch.nn.Module):  # type: ignore

    def __init__(
            self,
            mixture_size: int,
            do_layer_norm: bool = False,
            initial_scalar_parameters: Optional[List[float]] = None,
            trainable: bool = True,
    ) -> None:
        super().__init__()
        self.mixture_size = mixture_size
        self.do_layer_norm = do_layer_norm

        if initial_scalar_parameters is None:
            initial_scalar_parameters = [1.0 / mixture_size] * mixture_size
        elif len(initial_scalar_parameters) != mixture_size:
            raise ValueError("initial_scalar_parameters & mixture_size not match.")

        self.scalar_parameters = ParameterList([
                Parameter(
                        torch.FloatTensor([val]),
                        requires_grad=trainable,
                ) for val in initial_scalar_parameters
        ])
        self.gamma = Parameter(torch.FloatTensor([1.0]), requires_grad=trainable)

    def forward(
            self,
            tensors: List[torch.Tensor],  # pylint: disable=arguments-differ
            mask: torch.Tensor = None,
    ) -> torch.Tensor:

        def apply_layer_norm(tensor, broadcast_mask, num_elements_not_masked):  # type: ignore
            tensor_masked = tensor * broadcast_mask
            mean = torch.sum(tensor_masked) / num_elements_not_masked
            variance = torch.sum(torch.pow(
                    (tensor_masked - mean) * broadcast_mask,
                    2,
            )) / num_elements_not_masked
            return (tensor - mean) / torch.sqrt(variance + 1E-12)

        if len(tensors) != self.mixture_size:
            raise ValueError("tensors & mixture_size not match.")
        if self.do_layer_norm and mask is None:
            if tensors[0].ndimension() == 2:
                mask = torch.ones((tensors[0].shape[0],))
            else:
                raise ValueError("do_layer_norm but mask is not defined.")

        normed_weights = torch.split(
                torch.softmax(
                        torch.cat(list(self.scalar_parameters)),
                        0,
                ),
                1,
        )
        if self.do_layer_norm:
            mask_float = mask.float()
            broadcast_mask = mask_float.unsqueeze(-1)
            input_dim = tensors[0].size(-1)
            num_elements_not_masked = torch.sum(mask_float) * input_dim

        pieces = []
        for idx in range(self.mixture_size):
            tensor = tensors[idx]
            if self.do_layer_norm:
                tensor = apply_layer_norm(  # type: ignore
                        tensor,
                        broadcast_mask,
                        num_elements_not_masked,
                )
            weighted_tensor = normed_weights[idx] * tensor
            pieces.append(weighted_tensor)

        return self.gamma * sum(pieces)


class FastElmoBase(torch.nn.Module):  # type: ignore

    SCALAR_MIX_PARAMS = {
            'disable_scalar_mix',
            'num_output_representations',
            'output_representation_dropout',
            'scalar_mix_parameters',
            'do_layer_norm',
    }
    EXEC_PARAMS = {
            'exec_managed_lstm_bos_eos',
            'exec_managed_lstm_reset_states',
            'exec_sort_batch',
    }
    COMMON_PARAMS = SCALAR_MIX_PARAMS | EXEC_PARAMS

    _CHAR_CNN_FILTERS = [
            (1, 32),
            (2, 32),
            (3, 64),
            (4, 128),
            (5, 256),
            (6, 512),
            (7, 1024),
    ]

    def __init__(  # pylint: disable=dangerous-default-value
            self,

            # Generated by bilm-tf.
            options_file: Optional[str],
            weight_file: Optional[str],

            # Controls the behavior of execution.
            exec_managed_lstm_bos_eos: bool = True,
            exec_managed_lstm_reset_states: bool = False,
            exec_sort_batch: bool = True,

            # Controls the behavior of `ScalarMix`.
            disable_scalar_mix: bool = False,
            num_output_representations: int = 1,
            output_representation_dropout: float = 0.0,
            scalar_mix_parameters: Optional[List[float]] = None,
            do_layer_norm: bool = False,

            # Controls the behavior of factories.
            # Char CNN.
            disable_char_cnn: bool = False,
            char_cnn_requires_grad: bool = False,
            # From scratch.
            char_cnn_char_embedding_cnt: int = 261,
            char_cnn_char_embedding_dim: int = 16,
            char_cnn_filters: List[Tuple[int, int]] = _CHAR_CNN_FILTERS,
            char_cnn_activation: str = 'relu',
            char_cnn_num_highway_layers: int = 2,
            char_cnn_output_dim: int = 512,

            # Word Embedding.
            disable_word_embedding: bool = True,
            word_embedding_weight_file: Optional[str] = None,
            word_embedding_requires_grad: bool = False,
            # From scratch.
            word_embedding_cnt: int = 0,
            word_embedding_dim: int = 512,

            # The Forward LSTM.
            disable_forward_lstm: bool = False,
            forward_lstm_requires_grad: bool = False,
            # The Backward LSTM.
            disable_backward_lstm: bool = False,
            backward_lstm_requires_grad: bool = False,
            # From scratch.
            lstm_num_layers: int = 2,
            lstm_input_size: int = 512,
            lstm_hidden_size: int = 512,
            lstm_cell_size: int = 4096,
            lstm_cell_clip: float = 3.0,
            lstm_proj_clip: float = 3.0,
            lstm_truncated_bptt: int = 20,
            # Provide the BOS/EOS representations of shape `(projection_dim,)`
            # if char CNN is disabled.
            lstm_bos_repr: Optional[torch.Tensor] = None,
            lstm_eos_repr: Optional[torch.Tensor] = None,

            # The final softmax layer.
            disable_vocab_projection: bool = True,
            vocab_projection_requires_grad: bool = False,
            vocab_projection_input_size: int = 0,
            vocab_projection_proj_size: int = 0,
    ) -> None:
        super().__init__()

        self.disable_char_cnn = disable_char_cnn
        self.disable_word_embedding = disable_word_embedding
        self.disable_forward_lstm = disable_forward_lstm
        self.disable_backward_lstm = disable_backward_lstm
        self.disable_scalar_mix = disable_scalar_mix
        self.disable_vocab_projection = disable_vocab_projection

        self.exec_managed_lstm_bos_eos = exec_managed_lstm_bos_eos
        self.exec_managed_lstm_reset_states = exec_managed_lstm_reset_states
        self.exec_sort_batch = exec_sort_batch

        # Char CNN.
        if options_file:
            self.char_cnn_factory = ElmoCharacterEncoderFactory(
                    options_file,
                    weight_file,
            )
        else:
            # From scratch.
            self.char_cnn_factory = ElmoCharacterEncoderFactory.from_scratch(
                    char_embedding_cnt=char_cnn_char_embedding_cnt,
                    char_embedding_dim=char_cnn_char_embedding_dim,
                    filters=char_cnn_filters,
                    activation=char_cnn_activation,
                    num_highway_layers=char_cnn_num_highway_layers,
                    output_dim=char_cnn_output_dim,
            )

        if not disable_char_cnn:
            self._add_cpp_module_to_buffer(
                    'char_cnn',
                    self.char_cnn_factory.create(requires_grad=char_cnn_requires_grad),
            )

        # Word Embedding.
        if options_file:
            self.word_embedding_factory = ElmoWordEmbeddingFactory(
                    options_file,
                    word_embedding_weight_file or weight_file,
            )
        else:
            # From scratch.
            self.word_embedding_factory = ElmoWordEmbeddingFactory.from_scratch(
                    cnt=word_embedding_cnt,
                    dim=word_embedding_dim,
            )
            if exec_managed_lstm_bos_eos:
                raise ValueError('exec_managed_lstm_bos_eos should be disabled.')

        if not disable_word_embedding:
            # Not a cpp extension.
            word_embedding_weight, lstm_bos_repr, lstm_eos_repr = \
                    self.word_embedding_factory.create(requires_grad=word_embedding_requires_grad)
            self.register_buffer('word_embedding_weight', word_embedding_weight)

        # LSTM.
        if options_file:
            self.lstm_factory = ElmoLstmFactory(
                    options_file,
                    weight_file,
            )
        else:
            # From scratch.
            self.lstm_factory = ElmoLstmFactory.from_scratch(
                    num_layers=lstm_num_layers,
                    input_size=lstm_input_size,
                    hidden_size=lstm_hidden_size,
                    cell_size=lstm_cell_size,
                    cell_clip=lstm_cell_clip,
                    proj_clip=lstm_proj_clip,
                    truncated_bptt=lstm_truncated_bptt,
            )

        if not (disable_forward_lstm and disable_backward_lstm):
            forward_lstm, backward_lstm = self.lstm_factory.create(
                    enable_forward=not disable_forward_lstm,
                    forward_requires_grad=forward_lstm_requires_grad,
                    enable_backward=not disable_backward_lstm,
                    backward_requires_grad=backward_lstm_requires_grad,
            )
            if not disable_forward_lstm:
                self._add_cpp_module_to_buffer('forward_lstm', forward_lstm)
            if not disable_backward_lstm:
                self._add_cpp_module_to_buffer('backward_lstm', backward_lstm)

            # Cache BOS/EOS reprs.
            if exec_managed_lstm_bos_eos:
                if disable_char_cnn:
                    if lstm_bos_repr is None or lstm_eos_repr is None:
                        raise ValueError('BOS/EOS not provided.')

                else:
                    lstm_bos_repr, lstm_eos_repr = utils.get_bos_eos_token_repr(
                            self.char_cnn_factory,
                            self.char_cnn,
                    )
                self.register_buffer(
                        'lstm_bos_repr',
                        lstm_bos_repr,
                )
                self.register_buffer(
                        'lstm_eos_repr',
                        lstm_eos_repr,
                )

        # Vocabulary projection.
        if options_file:
            self.vocab_projection_factory = ElmoVocabProjectionFactory(
                    options_file,
                    weight_file,
            )
        else:
            self.vocab_projection_factory = ElmoVocabProjectionFactory.from_scratch(
                    vocab_projection_input_size,
                    vocab_projection_proj_size,
            )

        if not disable_vocab_projection:
            self.vocab_projection_weight, self.vocab_projection_bias = \
                    self.vocab_projection_factory.create(requires_grad=vocab_projection_requires_grad)

        # ScalarMix
        if not disable_scalar_mix:
            self.scalar_mixes: List[ScalarMix] = []

            for idx in range(num_output_representations):
                scalar_mix = ScalarMix(
                        # char cnn + lstm.
                        self.lstm_factory.num_layers + 1,
                        do_layer_norm=do_layer_norm,
                        initial_scalar_parameters=scalar_mix_parameters,
                        trainable=not scalar_mix_parameters,
                )
                self.add_module(f'scalar_mix_{idx}', scalar_mix)
                self.scalar_mixes.append(scalar_mix)

            self.repr_dropout = None
            if output_representation_dropout > 0.0:
                self.repr_dropout = torch.nn.Dropout(p=output_representation_dropout)

    def state_dict(  # type: ignore
            self,
            destination=None,
            prefix='',
            keep_vars=False,
    ):
        tmp_buffers = self._buffers
        self._buffers = OrderedDict()  # type: ignore
        ret = super().state_dict(destination, prefix, keep_vars)
        self._buffers = tmp_buffers
        return ret

    def _load_from_state_dict(  # type: ignore
            self,
            state_dict,
            prefix,
            local_metadata,
            strict,
            missing_keys,
            unexpected_keys,
            error_msgs,
    ):
        tmp_buffers = self._buffers
        self._buffers = OrderedDict()
        super()._load_from_state_dict(state_dict, prefix, local_metadata, strict, missing_keys,
                                      unexpected_keys, error_msgs)
        self._buffers = tmp_buffers

    def _add_cpp_module_to_buffer(self, name: str, cpp_module: Any) -> None:
        # register_buffer will raise an exception.
        self._buffers[name] = cpp_module

    def _get_lstm_device(self) -> int:
        cpp_ext = None
        if not self.disable_forward_lstm:
            cpp_ext = self.forward_lstm
        elif not self.disable_backward_lstm:
            cpp_ext = self.backward_lstm

        # Assume `cpp_ext` is not None.
        assert cpp_ext is not None
        tensor = cpp_ext.parameters()[0]
        return -1 if not tensor.is_cuda else tensor.get_device()  # type: ignore

    def get_batched_lstm_bos_eos_repr(self, attr_name: str, batch_size: int) -> PackedSequence:
        tensor = getattr(self, attr_name)

        if not tensor.is_cuda:
            # Move to GPU permanently.
            device = self._get_lstm_device()
            if device >= 0:
                tensor = tensor.cuda(device)
                setattr(self, attr_name, tensor)

        batched = tensor.unsqueeze(0).expand(batch_size, -1)
        return PackedSequence(batched, torch.LongTensor([batch_size]))

    def exec_forward_backword_lstm_bos_eos(
            self,
            lstm_attr_name: str,
            bos_eos_attr_name: str,
            batch_size: int,
    ) -> torch.Tensor:
        lstm = getattr(self, lstm_attr_name)
        batched = self.get_batched_lstm_bos_eos_repr(bos_eos_attr_name, batch_size)
        with torch.no_grad():
            outputs, _ = lstm(batched.data, batched.batch_sizes)
            # Returns the output of last layer.
            return outputs[-1]

    def exec_forward_lstm_bos(self, batch_size: int) -> torch.Tensor:
        return self.exec_forward_backword_lstm_bos_eos('forward_lstm', 'lstm_bos_repr', batch_size)

    def exec_forward_lstm_eos(self, batch_size: int) -> torch.Tensor:
        return self.exec_forward_backword_lstm_bos_eos('forward_lstm', 'lstm_eos_repr', batch_size)

    def exec_backward_lstm_bos(self, batch_size: int) -> torch.Tensor:
        return self.exec_forward_backword_lstm_bos_eos('backward_lstm', 'lstm_bos_repr', batch_size)

    def exec_backward_lstm_eos(self, batch_size: int) -> torch.Tensor:
        return self.exec_forward_backword_lstm_bos_eos('backward_lstm', 'lstm_eos_repr', batch_size)

    def exec_forward_lstm_permutate_states(self, index: torch.Tensor) -> None:
        self.forward_lstm.permutate_states(index)

    def exec_backward_lstm_permutate_states(self, index: torch.Tensor) -> None:
        self.backward_lstm.permutate_states(index)

    def exec_bilstm_permutate_states(self, index: torch.Tensor) -> None:
        if not self.disable_forward_lstm:
            self.exec_forward_lstm_permutate_states(index)
        if not self.disable_backward_lstm:
            self.exec_backward_lstm_permutate_states(index)

    def exec_char_cnn(self, inputs: PackedSequence) -> PackedSequence:
        """
        Char CNN.
        """
        output_data = self.char_cnn(inputs.data)
        return PackedSequence(output_data, inputs.batch_sizes)

    def exec_word_embedding(self, inputs: PackedSequence) -> PackedSequence:
        """
        Word embedding.
        """
        output_data = torch.nn.functional.embedding(
                inputs.data,
                self.word_embedding_weight,
                padding_idx=0,
        )
        return PackedSequence(output_data, inputs.batch_sizes)

    def exec_forward_lstm(
            self,
            inputs: PackedSequence,
    ) -> List[PackedSequence]:
        """
        Forward LSTM.
        """
        if self.exec_managed_lstm_bos_eos:
            max_batch_size = int(inputs.batch_sizes.data[0])
            # BOS.
            self.exec_forward_lstm_bos(max_batch_size)
        elif self.exec_managed_lstm_reset_states:
            self.forward_lstm.reset_states()

        # Feed inputs.
        outputs, _ = self.forward_lstm(inputs.data, inputs.batch_sizes)

        if self.exec_managed_lstm_bos_eos:
            # EOS.
            self.exec_forward_lstm_eos(max_batch_size)

        # To list of `PackedSequence`.
        return [PackedSequence(output, inputs.batch_sizes) for output in outputs]

    def exec_backward_lstm(
            self,
            inputs: PackedSequence,
    ) -> List[PackedSequence]:
        """
        Backward LSTM.
        """
        if self.exec_managed_lstm_bos_eos:
            max_batch_size = int(inputs.batch_sizes.data[0])
            # EOS.
            self.exec_backward_lstm_eos(max_batch_size)
        elif self.exec_managed_lstm_reset_states:
            self.backward_lstm.reset_states()

        # Feed inputs.
        outputs, _ = self.backward_lstm(inputs.data, inputs.batch_sizes)

        if self.exec_managed_lstm_bos_eos:
            # BOS.
            self.exec_backward_lstm_bos(max_batch_size)

        # To list of `PackedSequence`.
        return [PackedSequence(output, inputs.batch_sizes) for output in outputs]

    def exec_bilstm(
            self,
            inputs: PackedSequence,
    ) -> List[Tuple[PackedSequence, PackedSequence]]:
        """
        BiLSTM.
        """
        forward_seqs = self.exec_forward_lstm(inputs)
        backward_seqs = self.exec_backward_lstm(inputs)

        return list(zip(forward_seqs, backward_seqs))

    def concat_packed_sequences(
            self,
            packed_sequences: List[Tuple[PackedSequence, PackedSequence]],
    ) -> List[PackedSequence]:
        """
        Concatenate the outputs of fwd/bwd lstms.
        """
        return [
                PackedSequence(
                        torch.cat([fwd.data, bwd.data], dim=-1),
                        fwd.batch_sizes,
                ) for fwd, bwd in packed_sequences
        ]

    def combine_char_cnn_and_bilstm_outputs(
            self,
            char_cnn_packed: PackedSequence,
            bilstm_packed: List[PackedSequence],
    ) -> List[PackedSequence]:
        """
        Combine the outputs of Char CNN & BiLSTM for scalar mix.
        """
        # Simply duplicate the output of char cnn.
        duplicated_char_cnn_packed = PackedSequence(
                torch.cat([char_cnn_packed.data, char_cnn_packed.data], dim=-1),
                char_cnn_packed.batch_sizes,
        )

        combined = [duplicated_char_cnn_packed]
        combined.extend(bilstm_packed)
        return combined

    def exec_vocab_projection(self, context_repr: PackedSequence) -> PackedSequence:
        """
        Transform the last layer of LSTM to the probability distributions of vocabulary.
        """
        vocab_linear = torch.nn.functional.linear(
                context_repr.data,
                self.vocab_projection_weight,
                self.vocab_projection_bias,
        )
        vocab_probs = torch.nn.functional.softmax(vocab_linear, dim=-1)
        return PackedSequence(vocab_probs, context_repr.batch_sizes)

    def exec_scalar_mix(self, packed_sequences: List[PackedSequence]) -> List[PackedSequence]:
        """
        Scalar Mix.
        """
        reprs = []
        for scalar_mix in self.scalar_mixes:
            mixed = scalar_mix([inputs.data for inputs in packed_sequences])
            if self.repr_dropout is not None:
                mixed = self.repr_dropout(mixed)
            reprs.append(PackedSequence(mixed, packed_sequences[0].batch_sizes))
        return reprs

    def exec_bilstm_and_scalar_mix(
            self,
            token_repr: PackedSequence,
    ) -> List[PackedSequence]:
        """
        Common combination.
        """
        # BiLSTM.
        bilstm_repr = self.exec_bilstm(token_repr)
        # Scalar Mix.
        conbimed_repr = self.combine_char_cnn_and_bilstm_outputs(
                token_repr,
                self.concat_packed_sequences(bilstm_repr),
        )
        mixed_reprs = self.exec_scalar_mix(conbimed_repr)
        return mixed_reprs

    def pack_inputs(
            self,
            inputs: torch.Tensor,
            lengths: Optional[torch.Tensor] = None,
    ) -> PackedSequence:
        return utils.pack_inputs(inputs, lengths=lengths)

    def unpack_output(
            self,
            output: PackedSequence,
    ) -> torch.Tensor:
        return utils.unpack_outputs(output)

    def unpack_outputs(
            self,
            mixed_reprs: List[PackedSequence],
    ) -> List[torch.Tensor]:
        """
        Unpack the outputs of scalar mixtures.
        """
        return [self.unpack_output(mixed_repr) for mixed_repr in mixed_reprs]

    def to_allennlp_elmo_output_format(
            self,
            unpacks: List[torch.Tensor],
            mask: torch.Tensor,
    ) -> Dict[str, Union[torch.Tensor, List[torch.Tensor]]]:
        return {'elmo_representations': unpacks, 'mask': mask}

    def preprocess_inputs(
            self,
            inputs: torch.Tensor,
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor, Optional[torch.Tensor]]:
        lengths = utils.get_lengths_of_zero_padded_batch(inputs)
        original_lengths = lengths
        restoration_index: Optional[torch.Tensor] = None

        if self.exec_sort_batch:
            inputs, permutation_index, restoration_index = \
                    utils.sort_batch_by_length(inputs, lengths)
            lengths = lengths.index_select(0, permutation_index)
            self.exec_bilstm_permutate_states(permutation_index)

        return inputs, lengths, original_lengths, restoration_index

    def postprocess_outputs(
            self,
            unpacked_tensors: List[torch.Tensor],
            restoration_index: Optional[torch.Tensor],
            inputs: torch.Tensor,
            original_lengths: torch.Tensor,
    ) -> Tuple[List[torch.Tensor], torch.Tensor]:
        mask = utils.generate_mask_from_lengths(
                inputs.shape[0],
                inputs.shape[1],
                original_lengths,
        )
        if self.exec_sort_batch:
            assert restoration_index is not None
            unpacked_tensors = [
                    tensor.index_select(0, restoration_index) for tensor in unpacked_tensors
            ]
            self.exec_bilstm_permutate_states(restoration_index)

        return unpacked_tensors, mask

    def forward_with_sorting_and_packing(
            self,
            inputs: torch.Tensor,
    ) -> Tuple[List[torch.Tensor], torch.Tensor]:
        inputs, lengths, original_lengths, restoration_index = \
                self.preprocess_inputs(inputs)

        packed_inputs = self.pack_inputs(inputs, lengths)
        packed_outputs = self.execute(packed_inputs)

        unpacked_outputs = self.unpack_outputs(packed_outputs)
        unpacked_outputs, mask = self.postprocess_outputs(
                unpacked_outputs,
                restoration_index,
                inputs,
                original_lengths,
        )
        return unpacked_outputs, mask

    def execute(self, inputs: PackedSequence) -> List[PackedSequence]:
        raise NotImplementedError()

    def forward_like_allennlp(
            self,
            inputs: torch.Tensor,
    ) -> Dict[str, Union[torch.Tensor, List[torch.Tensor]]]:
        outputs, mask = self.forward_with_sorting_and_packing(inputs)
        return self.to_allennlp_elmo_output_format(outputs, mask)

    def forward(self):  # type: ignore
        raise NotImplementedError()


class FastElmo(FastElmoBase):

    def __init__(
            self,
            options_file: Optional[str],
            weight_file: str,
            **kwargs: Any,
    ) -> None:
        _raise_if_kwargs_is_invalid(
                self.COMMON_PARAMS | set([
                        # Fine-tuning is not fully supported by pytorch.
                        # 'char_cnn_requires_grad',
                        # 'forward_lstm_requires_grad',
                        # 'backward_lstm_requires_grad',
                ]),
                kwargs)
        super().__init__(options_file, weight_file, **kwargs)

    def execute(self, inputs: PackedSequence) -> List[PackedSequence]:
        token_repr = self.exec_char_cnn(inputs)
        mixed_reprs = self.exec_bilstm_and_scalar_mix(token_repr)
        return mixed_reprs

    def forward(  # type: ignore
            self,
            inputs: torch.Tensor,
    ) -> Dict[str, Union[torch.Tensor, List[torch.Tensor]]]:
        """
        The default workflow (same as AllenNLP).

        `inputs` of shape `(batch_size, max_timesteps, max_characters_per_token)
        """
        return self.forward_like_allennlp(inputs)


class FastElmoWordEmbedding(FastElmoBase):

    def __init__(
            self,
            options_file: Optional[str],
            weight_file: str,
            **kwargs: Any,
    ) -> None:
        _raise_if_kwargs_is_invalid(
                self.COMMON_PARAMS | {
                        'word_embedding_weight_file',
                        # Fine-tuning is not fully supported by pytorch.
                        # 'word_embedding_requires_grad',
                        # 'forward_lstm_requires_grad',
                        # 'backward_lstm_requires_grad',
                },
                kwargs)

        kwargs['disable_char_cnn'] = True
        kwargs['disable_word_embedding'] = False
        super().__init__(options_file, weight_file, **kwargs)

    def execute(self, inputs: PackedSequence) -> List[PackedSequence]:
        token_repr = self.exec_word_embedding(inputs)
        mixed_reprs = self.exec_bilstm_and_scalar_mix(token_repr)
        return mixed_reprs

    def forward(  # type: ignore
            self,
            inputs: torch.Tensor,
    ) -> Dict[str, Union[torch.Tensor, List[torch.Tensor]]]:
        """
        `inputs` of shape `(batch_size, max_timesteps)
        """
        return self.forward_like_allennlp(inputs)


class FastElmoPlainEncoderBase(FastElmoBase):  # pylint: disable=abstract-method

    def exec_context_independent_repr(self, inputs: PackedSequence) -> PackedSequence:
        raise NotImplementedError()

    def execute(self, inputs: PackedSequence) -> List[PackedSequence]:
        token_repr = self.exec_context_independent_repr(inputs)
        # BiLSTM.
        bilstm_repr = self.exec_bilstm(token_repr)
        # Scalar Mix.
        conbimed_repr = self.combine_char_cnn_and_bilstm_outputs(
                token_repr,
                self.concat_packed_sequences(bilstm_repr),
        )
        return conbimed_repr

    def forward(  # type: ignore
            self,
            inputs: torch.Tensor,
    ) -> Tuple[List[torch.Tensor], torch.Tensor]:
        """
        No scalar mix.
        """
        return self.forward_with_sorting_and_packing(inputs)


class FastElmoPlainEncoder(FastElmoPlainEncoderBase):

    def __init__(
            self,
            options_file: Optional[str],
            weight_file: str,
            **kwargs: Any,
    ) -> None:
        _raise_if_kwargs_is_invalid(self.EXEC_PARAMS, kwargs)
        kwargs['disable_scalar_mix'] = True
        super().__init__(options_file, weight_file, **kwargs)

    def exec_context_independent_repr(self, inputs: PackedSequence) -> PackedSequence:
        return self.exec_char_cnn(inputs)


class FastElmoWordEmbeddingPlainEncoder(FastElmoPlainEncoderBase):

    def __init__(
            self,
            options_file: Optional[str],
            weight_file: str,
            **kwargs: Any,
    ) -> None:
        _raise_if_kwargs_is_invalid(self.EXEC_PARAMS, kwargs)
        kwargs['disable_char_cnn'] = True
        kwargs['disable_word_embedding'] = False
        kwargs['disable_scalar_mix'] = True
        super().__init__(options_file, weight_file, **kwargs)

    def exec_context_independent_repr(self, inputs: PackedSequence) -> PackedSequence:
        return self.exec_word_embedding(inputs)


class FastElmoUnidirectionalVocabDistribBase(FastElmoBase):  # pylint: disable=abstract-method

    def exec_forward_vocab_prob_distrib(self, token_repr: PackedSequence) -> List[PackedSequence]:
        fwd_lstm_last = self.exec_forward_lstm(token_repr)[-1]
        fwd_vocab_distrib = self.exec_vocab_projection(fwd_lstm_last)
        return [fwd_vocab_distrib]

    def exec_backward_vocab_prob_distrib(self, token_repr: PackedSequence) -> List[PackedSequence]:
        bwd_lstm_last = self.exec_backward_lstm(token_repr)[-1]
        bwd_vocab_distrib = self.exec_vocab_projection(bwd_lstm_last)
        return [bwd_vocab_distrib]

    def forward(  # type: ignore
            self,
            inputs: torch.Tensor,
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        (vocab_distrib,), mask = self.forward_with_sorting_and_packing(inputs)
        return vocab_distrib, mask


class FastElmoForwardVocabDistrib(FastElmoUnidirectionalVocabDistribBase):

    def __init__(
            self,
            options_file: Optional[str],
            weight_file: str,
            **kwargs: Any,
    ) -> None:
        _raise_if_kwargs_is_invalid(
                {'exec_managed_lstm_bos_eos'},
                kwargs,
        )
        kwargs['disable_backward_lstm'] = True
        kwargs['disable_vocab_projection'] = False
        kwargs['disable_scalar_mix'] = True
        super().__init__(options_file, weight_file, **kwargs)

    def execute(self, inputs: PackedSequence) -> List[PackedSequence]:
        token_repr = self.exec_char_cnn(inputs)
        return self.exec_forward_vocab_prob_distrib(token_repr)


class FastElmoBackwardVocabDistrib(FastElmoUnidirectionalVocabDistribBase):

    def __init__(
            self,
            options_file: Optional[str],
            weight_file: str,
            **kwargs: Any,
    ) -> None:
        _raise_if_kwargs_is_invalid(
                {'exec_managed_lstm_bos_eos'},
                kwargs,
        )
        kwargs['disable_forward_lstm'] = True
        kwargs['disable_vocab_projection'] = False
        kwargs['disable_scalar_mix'] = True
        super().__init__(options_file, weight_file, **kwargs)

    def execute(self, inputs: PackedSequence) -> List[PackedSequence]:
        token_repr = self.exec_char_cnn(inputs)
        return self.exec_backward_vocab_prob_distrib(token_repr)


class FastElmoWordEmbeddingForwardVocabDistrib(FastElmoUnidirectionalVocabDistribBase):

    def __init__(
            self,
            options_file: Optional[str],
            weight_file: str,
            **kwargs: Any,
    ) -> None:
        _raise_if_kwargs_is_invalid(
                {'exec_managed_lstm_bos_eos', 'word_embedding_weight_file'},
                kwargs,
        )
        kwargs['disable_char_cnn'] = True
        kwargs['disable_word_embedding'] = False
        kwargs['disable_backward_lstm'] = True
        kwargs['disable_vocab_projection'] = False
        kwargs['disable_scalar_mix'] = True
        super().__init__(options_file, weight_file, **kwargs)

    def execute(self, inputs: PackedSequence) -> List[PackedSequence]:
        token_repr = self.exec_word_embedding(inputs)
        return self.exec_forward_vocab_prob_distrib(token_repr)


class FastElmoWordEmbeddingBackwardVocabDistrib(FastElmoUnidirectionalVocabDistribBase):

    def __init__(
            self,
            options_file: Optional[str],
            weight_file: str,
            **kwargs: Any,
    ) -> None:
        _raise_if_kwargs_is_invalid(
                {'exec_managed_lstm_bos_eos', 'word_embedding_weight_file'},
                kwargs,
        )
        kwargs['disable_char_cnn'] = True
        kwargs['disable_word_embedding'] = False
        kwargs['disable_forward_lstm'] = True
        kwargs['disable_vocab_projection'] = False
        kwargs['disable_scalar_mix'] = True
        super().__init__(options_file, weight_file, **kwargs)

    def execute(self, inputs: PackedSequence) -> List[PackedSequence]:
        token_repr = self.exec_word_embedding(inputs)
        return self.exec_backward_vocab_prob_distrib(token_repr)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup, find_packages
from torch.utils.cpp_extension import CppExtension, BuildExtension
from os.path import abspath, dirname


def load_requirements(path):
    with open(path) as fin:
        return [
            line
            for line in map(lambda l: l.strip(), fin.readlines())
            if line and not line.startswith('#')
        ]


with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = load_requirements('requirements_prod.txt')
test_requirements = load_requirements('requirements_dev.txt')

setup(
    author="Hunt Zhan",
    author_email='huntzhan.dev@gmail.com',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
    description="None",
    install_requires=requirements,
    license="MIT license",
    long_description=readme + '\n\n' + history,
    include_package_data=True,
    keywords='pytorch_fast_elmo',
    name='pytorch_fast_elmo',
    packages=find_packages(),
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/cnt-dev/pytorch-fast-elmo',
    version='0.6.12',
    zip_safe=False,
    # Pytorch Cpp Extension.
    ext_modules=[
        CppExtension(
            '_pytorch_fast_elmo',
            [
                'extension/elmo_character_encoder.cc',
                'extension/scalar_mix.cc',
                'extension/bind.cc',
            ],
            include_dirs=[dirname(abspath(__file__))],
        ),
    ],
    cmdclass={'build_ext': BuildExtension},
    entry_points={'console_scripts': ['fast-elmo = pytorch_fast_elmo.tool.cli:main']},
)



"""
All tests are copied from AllenNLP.
"""
import torch
import pytest
import numpy
from pytorch_fast_elmo import ScalarMix


def test_scalar_mix_can_run_forward():
    mixture = ScalarMix(3)
    tensors = [torch.randn([3, 4, 5]) for _ in range(3)]
    for k in range(3):
        mixture.scalar_parameters[k].data[0] = 0.1 * (k + 1)
    mixture.gamma.data[0] = 0.5
    result = mixture(tensors)

    weights = [0.1, 0.2, 0.3]
    normed_weights = numpy.exp(weights) / numpy.sum(numpy.exp(weights))
    expected_result = sum(normed_weights[k] * tensors[k].data.numpy() for k in range(3))
    expected_result *= 0.5
    numpy.testing.assert_almost_equal(expected_result, result.data.numpy())


def test_scalar_mix_throws_error_on_incorrect_number_of_inputs():
    mixture = ScalarMix(3)
    tensors = [torch.randn([3, 4, 5]) for _ in range(5)]
    with pytest.raises(ValueError):
        _ = mixture(tensors)


def test_scalar_mix_throws_error_on_incorrect_initial_scalar_parameters_length():
    with pytest.raises(ValueError):
        ScalarMix(3, initial_scalar_parameters=[0.0, 0.0])


def test_scalar_mix_trainable_with_initial_scalar_parameters():
    initial_scalar_parameters = [1.0, 2.0, 3.0]
    mixture = ScalarMix(3, initial_scalar_parameters=initial_scalar_parameters, trainable=False)
    for i, scalar_mix_parameter in enumerate(mixture.scalar_parameters):
        assert scalar_mix_parameter.requires_grad is False
        assert scalar_mix_parameter.item() == initial_scalar_parameters[i]


def test_scalar_mix_layer_norm():
    mixture = ScalarMix(3, do_layer_norm='scalar_norm_reg')

    tensors = [torch.randn([3, 4, 5]) for _ in range(3)]
    numpy_mask = numpy.ones((3, 4), dtype='int32')
    numpy_mask[1, 2:] = 0
    mask = torch.from_numpy(numpy_mask)

    weights = [0.1, 0.2, 0.3]
    for k in range(3):
        mixture.scalar_parameters[k].data[0] = weights[k]
    mixture.gamma.data[0] = 0.5
    result = mixture(tensors, mask)

    normed_weights = numpy.exp(weights) / numpy.sum(numpy.exp(weights))
    expected_result = numpy.zeros((3, 4, 5))
    for k in range(3):
        mean = numpy.mean(tensors[k].data.numpy()[numpy_mask == 1])
        std = numpy.std(tensors[k].data.numpy()[numpy_mask == 1])
        normed_tensor = (tensors[k].data.numpy() - mean) / (std + 1E-12)
        expected_result += (normed_tensor * normed_weights[k])
    expected_result *= 0.5

    numpy.testing.assert_almost_equal(expected_result, result.data.numpy(), decimal=6)


'''
def access_scalar_parameters(mixture, k):
    return mixture.named_parameters()['scalar_' + str(k)]


def access_gamma(mixture):
    return mixture.named_parameters()['gamma']


def test_scalar_mix_can_run_forward():
    mixture = ScalarMix(3)
    tensors = [torch.randn([3, 4, 5]) for _ in range(3)]
    for k in range(3):
        access_scalar_parameters(mixture, k).data[0] = 0.1 * (k + 1)
    access_gamma(mixture).data[0] = 0.5
    result = mixture(tensors)

    weights = [0.1, 0.2, 0.3]
    normed_weights = numpy.exp(weights) / numpy.sum(numpy.exp(weights))
    expected_result = sum(normed_weights[k] * tensors[k].data.numpy() for k in range(3))
    expected_result *= 0.5
    numpy.testing.assert_almost_equal(expected_result, result.data.numpy())


def test_scalar_mix_throws_error_on_incorrect_number_of_inputs():
    mixture = ScalarMix(3)
    tensors = [torch.randn([3, 4, 5]) for _ in range(5)]
    with pytest.raises(ValueError):
        _ = mixture(tensors)


def test_scalar_mix_throws_error_on_incorrect_initial_scalar_parameters_length():
    with pytest.raises(ValueError):
        ScalarMix(3, initial_scalar_parameters=[0.0, 0.0])


def test_scalar_mix_trainable_with_initial_scalar_parameters():
    initial_scalar_parameters = [1.0, 2.0, 3.0]
    mixture = ScalarMix(3, initial_scalar_parameters=initial_scalar_parameters, trainable=False)
    for k, initial_scalar_parameter in enumerate(initial_scalar_parameters):
        scalar_mix_parameter = access_scalar_parameters(mixture, k)
        assert scalar_mix_parameter.requires_grad is False
        assert scalar_mix_parameter.item() == initial_scalar_parameter


def test_scalar_mix_layer_norm():
    mixture = ScalarMix(3, do_layer_norm=True)

    tensors = [torch.randn([3, 4, 5]) for _ in range(3)]
    numpy_mask = numpy.ones((3, 4), dtype='int32')
    numpy_mask[1, 2:] = 0
    mask = torch.from_numpy(numpy_mask)

    weights = [0.1, 0.2, 0.3]
    for k in range(3):
        access_scalar_parameters(mixture, k).data[0] = weights[k]
    access_gamma(mixture).data[0] = 0.5
    result = mixture(tensors, mask)

    normed_weights = numpy.exp(weights) / numpy.sum(numpy.exp(weights))
    expected_result = numpy.zeros((3, 4, 5))
    for k in range(3):
        mean = numpy.mean(tensors[k].data.numpy()[numpy_mask == 1])
        std = numpy.std(tensors[k].data.numpy()[numpy_mask == 1])
        normed_tensor = (tensors[k].data.numpy() - mean) / (std + 1E-12)
        expected_result += (normed_tensor * normed_weights[k])
    expected_result *= 0.5

    numpy.testing.assert_almost_equal(expected_result, result.data.numpy(), decimal=6)


def test_scalar_mix_layer_norm_packed_sequence():
    mixture = ScalarMix(3, do_layer_norm=True)
    tensors = [torch.randn([3, 4]) for _ in range(3)]
    mixture(tensors)
'''



from typing import List, Any, Tuple
import random
import string
import time
import statistics

import torch
from pytorch_fast_elmo import batch_to_char_ids, FastElmo


class SentenceGenerator:

    def __init__(
            self,
            word_min: int,
            word_max: int,
            sent_min: int,
            sent_max: int,
    ) -> None:
        self.word_min = word_min
        self.word_max = word_max
        self.sent_min = sent_min
        self.sent_max = sent_max

    def generate_sentence(self) -> List[str]:
        return [
                ''.join(
                        random.choices(
                                string.ascii_lowercase,
                                k=random.randint(self.word_min, self.word_max),
                        )) for _ in range(random.randint(self.sent_min, self.sent_max))
        ]

    def generate_batch(self, batch_size: int) -> List[List[str]]:
        return [self.generate_sentence() for _ in range(batch_size)]


def load_fast_elmo(
        options_file: str,
        weight_file: str,
) -> FastElmo:
    return FastElmo(
            options_file,
            weight_file,
            scalar_mix_parameters=[1.0, 1.0, 1.0],
    )


def load_allennlp_elmo(
        options_file: str,
        weight_file: str,
) -> Any:
    from allennlp.modules.elmo import Elmo
    return Elmo(
            options_file,
            weight_file,
            num_output_representations=1,
            dropout=0.0,
            scalar_mix_parameters=[1.0, 1.0, 1.0],
    )


def profile_full_elmo(
        mode: str,
        options_file: str,
        weight_file: str,
        cuda_device: int,
        cuda_synchronize: bool,
        batch_size: int,
        warmup_size: int,
        iteration_size: int,
        word_min: int,
        word_max: int,
        sent_min: int,
        sent_max: int,
        random_seed: int,
) -> Tuple[float, float, float]:
    random.seed(random_seed)

    module: Any = None
    if mode == 'fast-elmo':
        module = load_fast_elmo(options_file, weight_file)
    elif mode == 'allennlp-elmo':
        module = load_allennlp_elmo(options_file, weight_file)
    else:
        raise ValueError('invalid mode')

    sent_gen = SentenceGenerator(
            word_min,
            word_max,
            sent_min,
            sent_max,
    )

    if cuda_device >= 0:
        module = module.cuda(cuda_device)

    durations: List[float] = []

    for idx in range(warmup_size + iteration_size):
        batch = sent_gen.generate_batch(batch_size)
        char_ids = batch_to_char_ids(batch)

        if cuda_device >= 0:
            char_ids = char_ids.cuda(cuda_device)
            if cuda_synchronize:
                torch.cuda.synchronize()

        start = time.time()
        with torch.no_grad():
            module(char_ids)
        if cuda_device >= 0 and cuda_synchronize:
            torch.cuda.synchronize()
        end = time.time()

        if idx >= warmup_size:
            durations.append(end - start)

    mean = statistics.mean(durations)
    median = statistics.median(durations)
    stdev = statistics.stdev(durations)

    return mean, median, stdev



# pylint: disable=no-self-use
import cProfile
import pstats
import io

import fire
from pytorch_fast_elmo import utils
from pytorch_fast_elmo.tool import profile, inspect


class Main:

    def cache_char_cnn(  # type: ignore
            self,
            vocab_txt,
            options_file,
            weight_file,
            txt_out,
            max_characters_per_token=utils.ElmoCharacterIdsConst.MAX_WORD_LENGTH,
            cuda_device=-1,
            batch_size=256,
    ):
        utils.cache_char_cnn_vocab(
                vocab_txt,
                options_file,
                weight_file,
                txt_out,
                max_characters_per_token,
                cuda_device,
                batch_size,
        )

    def export_word_embd(  # type: ignore
            self,
            vocab_txt,
            weight_file,
            txt_out,
    ):
        utils.export_word_embd(
                vocab_txt,
                weight_file,
                txt_out,
        )

    def profile_full(  # type: ignore
            self,
            mode,
            options_file,
            weight_file,
            cuda_device=-1,
            cuda_synchronize=False,
            batch_size=32,
            warmup_size=20,
            iteration_size=1000,
            word_min=1,
            word_max=20,
            sent_min=1,
            sent_max=30,
            random_seed=10000,
            profiler=False,
            output_file=None,
    ):
        sstream = io.StringIO()

        if profiler:
            cpr = cProfile.Profile()
            cpr.enable()

        mean, median, stdev = profile.profile_full_elmo(
                mode,
                options_file,
                weight_file,
                cuda_device,
                cuda_synchronize,
                batch_size,
                warmup_size,
                iteration_size,
                word_min,
                word_max,
                sent_min,
                sent_max,
                random_seed,
        )

        sstream.write(f'Finish {iteration_size} iterations.\n')
        sstream.write(f'Mode: {mode}\n')
        sstream.write(f'Duration Mean: {mean}\n')
        sstream.write(f'Duration Median: {median}\n')
        sstream.write(f'Duration Stdev: {stdev}\n\n')

        if profiler:
            cpr.disable()
            pstats.Stats(cpr, stream=sstream).sort_stats('cumulative').print_stats()

        if output_file:
            with open(output_file, 'w') as fout:
                fout.write(sstream.getvalue())
        else:
            print(sstream.getvalue())

    def sample_sentence(  # type: ignore
            self,
            options_file,
            weight_file,
            vocab_txt,
            output_json,
            enable_trace=False,
            go_forward=True,
            no_char_cnn=False,
            char_cnn_maxlen=0,
            next_token_top_k=5,
            sample_size=1,
            sample_constrain_txt=None,
            warm_up_txt=None,
            cuda_device=-1,
    ):
        inspect.sample_sentence(
                options_file,
                weight_file,
                vocab_txt,
                output_json,
                enable_trace,
                no_char_cnn,
                char_cnn_maxlen,
                go_forward,
                next_token_top_k,
                sample_size,
                sample_constrain_txt,
                warm_up_txt,
                cuda_device,
        )

    def encode_sentences(  # type: ignore
            self,
            options_file,
            weight_file,
            vocab_txt,
            input_txt,
            output_hdf5,
            no_char_cnn=False,
            char_cnn_maxlen=0,
            scalar_mix=None,
            warm_up_txt=None,
            cuda_device=-1,
    ):
        inspect.encode_sentences(
                options_file,
                weight_file,
                vocab_txt,
                input_txt,
                output_hdf5,
                no_char_cnn,
                char_cnn_maxlen,
                scalar_mix,
                warm_up_txt,
                cuda_device,
        )


def main():  # type: ignore
    fire.Fire(Main)
